from aioredis.util import wait_convert, wait_ok, _NOTSET, PY_35

if PY_35:
    from aioredis.util import _ScanIter


class GenericCommandsMixin:
    """Generic commands mixin.

    For commands details see: http://redis.io/commands/#generic
    """

    def delete(self, key, *keys):
        """Delete a key."""
        fut = self.execute(b'DEL', key, *keys)
        return wait_convert(fut, int)

    def dump(self, key):
        """Dump a key."""
        return self.execute(b'DUMP', key)

    def exists(self, key, *keys):
        """Check if key(s) exists.

        .. versionchanged:: v0.2.9
           Accept multiple keys; **return** type **changed** from bool to int.
        """
        return self.execute(b'EXISTS', key, *keys)

    def expire(self, key, timeout):
        """Set a timeout on key.

        if timeout is float it will be multiplied by 1000
        coerced to int and passed to `pexpire` method.

        Otherwise raises TypeError if timeout argument is not int.
        """
        if isinstance(timeout, float):
            return self.pexpire(key, int(timeout * 1000))
        if not isinstance(timeout, int):
            raise TypeError(
                "timeout argument must be int, not {!r}".format(timeout))
        fut = self.execute(b'EXPIRE', key, timeout)
        return wait_convert(fut, bool)

    def expireat(self, key, timestamp):
        """Set expire timestamp on a key.

        if timeout is float it will be multiplied by 1000
        coerced to int and passed to `pexpireat` method.

        Otherwise raises TypeError if timestamp argument is not int.
        """
        if isinstance(timestamp, float):
            return self.pexpireat(key, int(timestamp * 1000))
        if not isinstance(timestamp, int):
            raise TypeError("timestamp argument must be int, not {!r}"
                            .format(timestamp))
        fut = self.execute(b'EXPIREAT', key, timestamp)
        return wait_convert(fut, bool)

    def keys(self, pattern, *, encoding=_NOTSET):
        """Returns all keys matching pattern."""
        return self.execute(b'KEYS', pattern, encoding=encoding)

    def migrate(self, host, port, key, dest_db, timeout, *,
                copy=False, replace=False):
        """Atomically transfer a key from a Redis instance to another one."""
        if not isinstance(host, str):
            raise TypeError("host argument must be str")
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if not isinstance(dest_db, int):
            raise TypeError("dest_db argument must be int")
        if not host:
            raise ValueError("Got empty host")
        if dest_db < 0:
            raise ValueError("dest_db must be greater equal 0")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")

        flags = []
        if copy:
            flags.append(b'COPY')
        if replace:
            flags.append(b'REPLACE')
        fut = self.execute(b'MIGRATE', host, port,
                           key, dest_db, timeout, *flags)
        return wait_ok(fut)

    def migrate_keys(self, host, port, keys, dest_db, timeout, *,
                     copy=False, replace=False):
        """Atomically transfer keys from one Redis instance to another one.

        Keys argument must be list/tuple of keys to migrate.
        """
        if not isinstance(host, str):
            raise TypeError("host argument must be str")
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int")
        if not isinstance(dest_db, int):
            raise TypeError("dest_db argument must be int")
        if not isinstance(keys, (list, tuple)):
            raise TypeError("keys argument must be list or tuple")
        if not host:
            raise ValueError("Got empty host")
        if dest_db < 0:
            raise ValueError("dest_db must be greater equal 0")
        if timeout < 0:
            raise ValueError("timeout must be greater equal 0")
        if not keys:
            raise ValueError("keys must not be empty")

        flags = []
        if copy:
            flags.append(b'COPY')
        if replace:
            flags.append(b'REPLACE')
        flags.append(b'KEYS')
        flags.extend(keys)
        fut = self.execute(b'MIGRATE', host, port,
                           "", dest_db, timeout, *flags)
        return wait_ok(fut)

    def move(self, key, db):
        """Move key from currently selected database to specified destination.

        :raises TypeError: if db is not int
        :raises ValueError: if db is less then 0
        """
        if not isinstance(db, int):
            raise TypeError("db argument must be int, not {!r}".format(db))
        if db < 0:
            raise ValueError("db argument must be not less then 0, {!r}"
                             .format(db))
        fut = self.execute(b'MOVE', key, db)
        return wait_convert(fut, bool)

    def object_refcount(self, key):
        """Returns the number of references of the value associated
        with the specified key (OBJECT REFCOUNT).
        """
        return self.execute(b'OBJECT', b'REFCOUNT', key)

    def object_encoding(self, key):
        """Returns the kind of internal representation used in order
        to store the value associated with a key (OBJECT ENCODING).
        """
        # TODO: set default encoding to 'utf-8'
        return self.execute(b'OBJECT', b'ENCODING', key)

    def object_idletime(self, key):
        """Returns the number of seconds since the object is not requested
        by read or write operations (OBJECT IDLETIME).
        """
        return self.execute(b'OBJECT', b'IDLETIME', key)

    def persist(self, key):
        """Remove the existing timeout on key."""
        fut = self.execute(b'PERSIST', key)
        return wait_convert(fut, bool)

    def pexpire(self, key, timeout):
        """Set a milliseconds timeout on key.

        :raises TypeError: if timeout is not int
        """
        if not isinstance(timeout, int):
            raise TypeError("timeout argument must be int, not {!r}"
                            .format(timeout))
        fut = self.execute(b'PEXPIRE', key, timeout)
        return wait_convert(fut, bool)

    def pexpireat(self, key, timestamp):
        """Set expire timestamp on key, timestamp in milliseconds.

        :raises TypeError: if timeout is not int
        """
        if not isinstance(timestamp, int):
            raise TypeError("timestamp argument must be int, not {!r}"
                            .format(timestamp))
        fut = self.execute(b'PEXPIREAT', key, timestamp)
        return wait_convert(fut, bool)

    def pttl(self, key):
        """Returns time-to-live for a key, in milliseconds.

        Special return values (starting with Redis 2.8):

        * command returns -2 if the key does not exist.
        * command returns -1 if the key exists but has no associated expire.
        """
        # TODO: maybe convert negative values to:
        #       -2 to None  - no key
        #       -1 to False - no expire
        return self.execute(b'PTTL', key)

    def randomkey(self, *, encoding=_NOTSET):
        """Return a random key from the currently selected database."""
        return self.execute(b'RANDOMKEY', encoding=encoding)

    def rename(self, key, newkey):
        """Renames key to newkey.

        :raises ValueError: if key == newkey
        """
        if key == newkey:
            raise ValueError("key and newkey are the same")
        fut = self.execute(b'RENAME', key, newkey)
        return wait_ok(fut)

    def renamenx(self, key, newkey):
        """Renames key to newkey only if newkey does not exist.

        :raises ValueError: if key == newkey
        """
        if key == newkey:
            raise ValueError("key and newkey are the same")
        fut = self.execute(b'RENAMENX', key, newkey)
        return wait_convert(fut, bool)

    def restore(self, key, ttl, value):
        """Creates a key associated with a value that is obtained via DUMP."""
        return self.execute(b'RESTORE', key, ttl, value)

    def scan(self, cursor=0, match=None, count=None):
        """Incrementally iterate the keys space.

        Usage example:

        >>> match = 'something*'
        >>> cur = b'0'
        >>> while cur:
        ...     cur, keys = await redis.scan(cur, match=match)
        ...     for key in keys:
        ...         print('Matched:', key)

        """
        args = []
        if match is not None:
            args += [b'MATCH', match]
        if count is not None:
            args += [b'COUNT', count]
        fut = self.execute(b'SCAN', cursor, *args)
        return wait_convert(fut, lambda o: (int(o[0]), o[1]))

    if PY_35:
        def iscan(self, *, match=None, count=None):
            """Incrementally iterate the keys space using async for.

            Usage example:

            >>> async for key in redis.iscan(match='something*'):
            ...     print('Matched:', key)

            """
            return _ScanIter(lambda cur: self.scan(cur,
                                                   match=match, count=count))

    def sort(self, key, *get_patterns,
             by=None, offset=None, count=None,
             asc=None, alpha=False, store=None):
        """Sort the elements in a list, set or sorted set."""
        args = []
        if by is not None:
            args += [b'BY', by]
        if offset is not None and count is not None:
            args += [b'LIMIT', offset, count]
        if get_patterns:
            args += sum(([b'GET', pattern] for pattern in get_patterns), [])
        if asc is not None:
            args += [asc is True and b'ASC' or b'DESC']
        if alpha:
            args += [b'ALPHA']
        if store is not None:
            args += [b'STORE', store]
        return self.execute(b'SORT', key, *args)

    def ttl(self, key):
        """Returns time-to-live for a key, in seconds.

        Special return values (starting with Redis 2.8):
        * command returns -2 if the key does not exist.
        * command returns -1 if the key exists but has no associated expire.
        """
        # TODO: maybe convert negative values to:
        #       -2 to None  - no key
        #       -1 to False - no expire
        return self.execute(b'TTL', key)

    def type(self, key):
        """Returns the string representation of the value's type stored at key.
        """
        # NOTE: for non-existent keys TYPE returns b'none'
        return self.execute(b'TYPE', key)



import asyncio
import pytest

from aioredis import ReplyError


@asyncio.coroutine
def add(redis, key, field, value):
    ok = yield from redis.connection.execute(
        b'hset', key, field, value)
    assert ok == 1


@pytest.mark.run_loop
def test_hdel(redis):
    key, field, value = b'key:hdel', b'bar', b'zap'
    yield from add(redis, key, field, value)
    # delete value that exists, expected 1
    result = yield from redis.hdel(key, field)
    assert result == 1
    # delete value that does not exists, expected 0
    result = yield from redis.hdel(key, field)
    assert result == 0

    with pytest.raises(TypeError):
        yield from redis.hdel(None, field)


@pytest.mark.run_loop
def test_hexists(redis):
    key, field, value = b'key:hexists', b'bar', b'zap'
    yield from add(redis, key, field, value)
    # check value that exists, expected 1
    result = yield from redis.hexists(key, field)
    assert result == 1
    # check value when, key exists and field does not, expected 0
    result = yield from redis.hexists(key, b'not:' + field)
    assert result == 0
    # check value when, key not exists, expected 0
    result = yield from redis.hexists(b'not:' + key, field)
    assert result == 0

    with pytest.raises(TypeError):
        yield from redis.hexists(None, field)


@pytest.mark.run_loop
def test_hget(redis):

    key, field, value = b'key:hget', b'bar', b'zap'
    yield from add(redis, key, field, value)
    # basic test, fetch value and check in to reference
    test_value = yield from redis.hget(key, field)
    assert test_value == value
    # fetch value, when field does not exists
    test_value = yield from redis.hget(key, b'not' + field)
    assert test_value is None
    # fetch value when key does not exists
    test_value = yield from redis.hget(b'not:' + key, b'baz')
    assert test_value is None

    # check encoding
    test_value = yield from redis.hget(key, field, encoding='utf-8')
    assert test_value == 'zap'

    with pytest.raises(TypeError):
        yield from redis.hget(None, field)


@pytest.mark.run_loop
def test_hgetall(redis):
    yield from add(redis, 'key:hgetall', 'foo', 'baz')
    yield from add(redis, 'key:hgetall', 'bar', 'zap')

    test_value = yield from redis.hgetall('key:hgetall')
    assert isinstance(test_value, dict)
    assert {b'foo': b'baz', b'bar': b'zap'} == test_value
    # try to get all values from key that does not exits
    test_value = yield from redis.hgetall(b'not:key:hgetall')
    assert test_value == {}

    # check encoding param
    test_value = yield from redis.hgetall(
        'key:hgetall', encoding='utf-8')
    assert {'foo': 'baz', 'bar': 'zap'} == test_value

    with pytest.raises(TypeError):
        yield from redis.hgetall(None)


@pytest.mark.run_loop
def test_hincrby(redis):
    key, field, value = b'key:hincrby', b'bar', 1
    yield from add(redis, key, field, value)
    # increment initial value by 2
    result = yield from redis.hincrby(key, field, 2)
    assert result == 3

    result = yield from redis.hincrby(key, field, -1)
    assert result == 2

    result = yield from redis.hincrby(key, field, -100)
    assert result == -98

    result = yield from redis.hincrby(key, field, -2)
    assert result == -100

    # increment value in case of key or field that does not exists
    result = yield from redis.hincrby(b'not:' + key, field, 2)
    assert result == 2
    result = yield from redis.hincrby(key, b'not:' + field, 2)
    assert result == 2

    with pytest.raises(ReplyError):
        yield from redis.hincrby(key, b'not:' + field, 3.14)

    with pytest.raises(ReplyError):
        # initial value is float, try to increment 1
        yield from add(redis, b'other:' + key, field, 3.14)
        yield from redis.hincrby(b'other:' + key, field, 1)

    with pytest.raises(TypeError):
        yield from redis.hincrby(None, field, 2)


@pytest.mark.run_loop
def test_hincrbyfloat(redis):
    key, field, value = b'key:hincrbyfloat', b'bar', 2.71
    yield from add(redis, key, field, value)

    result = yield from redis.hincrbyfloat(key, field, 3.14)
    assert result == 5.85

    result = yield from redis.hincrbyfloat(key, field, -2.71)
    assert result == 3.14

    result = yield from redis.hincrbyfloat(key, field, -100.1)
    assert result == -96.96

    # increment value in case of key or field that does not exists
    result = yield from redis.hincrbyfloat(b'not:' + key, field, 3.14)
    assert result == 3.14

    result = yield from redis.hincrbyfloat(key, b'not:' + field, 3.14)
    assert result == 3.14

    with pytest.raises(TypeError):
        yield from redis.hincrbyfloat(None, field, 2)


@pytest.mark.run_loop
def test_hkeys(redis):
    key = b'key:hkeys'
    field1, field2 = b'foo', b'bar'
    value1, value2 = b'baz', b'zap'
    yield from add(redis, key, field1, value1)
    yield from add(redis, key, field2, value2)

    test_value = yield from redis.hkeys(key)
    assert set(test_value) == {field1, field2}

    test_value = yield from redis.hkeys(b'not:' + key)
    assert test_value == []

    test_value = yield from redis.hkeys(key, encoding='utf-8')
    assert set(test_value) == {'foo', 'bar'}

    with pytest.raises(TypeError):
        yield from redis.hkeys(None)


@pytest.mark.run_loop
def test_hlen(redis):
    key = b'key:hlen'
    field1, field2 = b'foo', b'bar'
    value1, value2 = b'baz', b'zap'
    yield from add(redis, key, field1, value1)
    yield from add(redis, key, field2, value2)

    test_value = yield from redis.hlen(key)
    assert test_value == 2

    test_value = yield from redis.hlen(b'not:' + key)
    assert test_value == 0

    with pytest.raises(TypeError):
        yield from redis.hlen(None)


@pytest.mark.run_loop
def test_hmget(redis):
    key = b'key:hmget'
    field1, field2 = b'foo', b'bar'
    value1, value2 = b'baz', b'zap'
    yield from add(redis, key, field1, value1)
    yield from add(redis, key, field2, value2)

    test_value = yield from redis.hmget(key, field1, field2)
    assert set(test_value) == {value1, value2}

    test_value = yield from redis.hmget(
        key, b'not:' + field1, b'not:' + field2)
    assert [None, None] == test_value

    val = yield from redis.hincrby(key, 'numeric')
    assert val == 1
    test_value = yield from redis.hmget(
        key, field1, field2, 'numeric', encoding='utf-8')
    assert ['baz', 'zap', '1'] == test_value

    with pytest.raises(TypeError):
        yield from redis.hmget(None, field1, field2)


@pytest.mark.run_loop
def test_hmset(redis):
    key, field, value = b'key:hmset', b'bar', b'zap'
    yield from add(redis, key, field, value)

    # key and field exists
    test_value = yield from redis.hmset(key, field, b'baz')
    assert test_value is True

    result = yield from redis.hexists(key, field)
    assert result == 1

    # key and field does not exists
    test_value = yield from redis.hmset(b'not:' + key, field, value)
    assert test_value is True
    result = yield from redis.hexists(b'not:' + key, field)
    assert result == 1

    # set multiple
    pairs = [b'foo', b'baz', b'bar', b'paz']
    test_value = yield from redis.hmset(key, *pairs)
    assert test_value is True
    test_value = yield from redis.hmget(key, b'foo', b'bar')
    assert set(test_value) == {b'baz', b'paz'}

    with pytest.raises(TypeError):
        yield from redis.hmset(key, b'foo', b'bar', b'baz')

    with pytest.raises(TypeError):
        yield from redis.hmset(None, *pairs)

    with pytest.raises(TypeError):
        yield from redis.hmset(key, {'foo': 'bar'}, {'baz': 'bad'})

    with pytest.raises(TypeError):
        yield from redis.hmset(key)


@pytest.mark.run_loop
def test_hmset_dict(redis):
    key = 'key:hmset'

    # dict
    d1 = {b'foo': b'one dict'}
    test_value = yield from redis.hmset_dict(key, d1)
    assert test_value is True
    test_value = yield from redis.hget(key, b'foo')
    assert test_value == b'one dict'

    # kwdict
    test_value = yield from redis.hmset_dict(key, foo=b'kw1', bar=b'kw2')
    assert test_value is True
    test_value = yield from redis.hmget(key, b'foo', b'bar')
    assert set(test_value) == {b'kw1', b'kw2'}

    # dict & kwdict
    d1 = {b'foo': b'dict'}
    test_value = yield from redis.hmset_dict(key, d1, foo=b'kw')
    assert test_value is True
    test_value = yield from redis.hget(key, b'foo')
    assert test_value == b'kw'

    # allow empty dict with kwargs
    test_value = yield from redis.hmset_dict(key, {}, foo='kw')
    assert test_value is True
    test_value = yield from redis.hget(key, 'foo')
    assert test_value == b'kw'

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key)

    with pytest.raises(ValueError):
        yield from redis.hmset_dict(key, {})

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key, ('foo', 'pairs'))

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key, b'foo', 'pairs')

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key, b'foo', 'pairs', foo=b'kw1')

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key, {'a': 1}, {'b': 2})

    with pytest.raises(TypeError):
        yield from redis.hmset_dict(key, {'a': 1}, {'b': 2}, 'c', 3, d=4)


@pytest.mark.run_loop
def test_hset(redis):
    key, field, value = b'key:hset', b'bar', b'zap'
    test_value = yield from redis.hset(key, field, value)
    assert test_value == 1

    test_value = yield from redis.hset(key, field, value)
    assert test_value == 0

    test_value = yield from redis.hset(b'other:' + key, field, value)
    assert test_value == 1

    result = yield from redis.hexists(b'other:' + key, field)
    assert result == 1

    with pytest.raises(TypeError):
        yield from redis.hset(None, field, value)


@pytest.mark.run_loop
def test_hsetnx(redis):
    key, field, value = b'key:hsetnx', b'bar', b'zap'
    # field does not exists, operation should be successful
    test_value = yield from redis.hsetnx(key, field, value)
    assert test_value == 1
    # make sure that value was stored
    result = yield from redis.hget(key, field)
    assert result == value
    # field exists, operation should not change any value
    test_value = yield from redis.hsetnx(key, field, b'baz')
    assert test_value == 0
    # make sure value was not changed
    result = yield from redis.hget(key, field)
    assert result == value

    with pytest.raises(TypeError):
        yield from redis.hsetnx(None, field, value)


@pytest.mark.run_loop
def test_hvals(redis):
    key = b'key:hvals'
    field1, field2 = b'foo', b'bar'
    value1, value2 = b'baz', b'zap'
    yield from add(redis, key, field1, value1)
    yield from add(redis, key, field2, value2)

    test_value = yield from redis.hvals(key)
    assert set(test_value) == {value1, value2}

    test_value = yield from redis.hvals(b'not:' + key)
    assert test_value == []

    test_value = yield from redis.hvals(key, encoding='utf-8')
    assert set(test_value) == {'baz', 'zap'}
    with pytest.raises(TypeError):
        yield from redis.hvals(None)


@pytest.redis_version(2, 8, 0, reason='HSCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
def test_hscan(redis):
    key = b'key:hscan'
    # setup initial values 3 "field:foo:*" items and 7 "field:bar:*" items
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        f = 'field:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        v = 'value:{}'.format(i).encode('utf-8')
        yield from add(redis, key, f, v)
    # fetch 'field:foo:*' items expected tuple with 3 fields and 3 values
    cursor, values = yield from redis.hscan(key, match=b'field:foo:*')
    assert len(values) == 3*2
    # fetch 'field:bar:*' items expected tuple with 7 fields and 7 values
    cursor, values = yield from redis.hscan(key, match=b'field:bar:*')
    assert len(values) == 7*2

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    cursor = b'0'
    test_values = []
    while cursor:
        cursor, values = yield from redis.hscan(key, cursor, count=1)
        test_values.extend(values)
    assert len(test_values) == 10*2

    with pytest.raises(TypeError):
        yield from redis.hscan(None)


@pytest.mark.run_loop
def test_hgetall_enc(create_redis, loop, server):
    redis = yield from create_redis(
        server.tcp_address, loop=loop, encoding='utf-8')
    TEST_KEY = 'my-key-nx'
    yield from redis.hmset(TEST_KEY, 'foo', 'bar', 'baz', 'bad')

    tr = redis.multi_exec()
    tr.hgetall(TEST_KEY)
    res = yield from tr.execute()
    assert res == [{'foo': 'bar', 'baz': 'bad'}]


@pytest.mark.run_loop
@pytest.redis_version(3, 2, 0, reason="HSTRLEN new in redis 3.2.0")
def test_hstrlen(redis):
    ok = yield from redis.hset('myhash', 'str_field', 'some value')
    assert ok == 1
    ok = yield from redis.hincrby('myhash', 'uint_field', 1)
    assert ok == 1

    ok = yield from redis.hincrby('myhash', 'int_field', -1)
    assert ok == -1

    l = yield from redis.hstrlen('myhash', 'str_field')
    assert l == 10
    l = yield from redis.hstrlen('myhash', 'uint_field')
    assert l == 1
    l = yield from redis.hstrlen('myhash', 'int_field')
    assert l == 2

    l = yield from redis.hstrlen('myhash', 'none_field')
    assert l == 0

    l = yield from redis.hstrlen('none_key', 'none_field')
    assert l == 0



import pytest


@pytest.redis_version(2, 8, 0, reason='HSCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_ihscan(redis):
    key = b'key:hscan'
    # setup initial values 3 "field:foo:*" items and 7 "field:bar:*" items
    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        f = 'field:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        v = 'value:{}'.format(i).encode('utf-8')
        assert await redis.hset(key, f, v) == 1

    async def coro(cmd):
        lst = []
        async for i in cmd:
            lst.append(i)
        return lst

    # fetch 'field:foo:*' items expected tuple with 3 fields and 3 values
    ret = await coro(redis.ihscan(key, match=b'field:foo:*'))
    assert set(ret) == {(b'field:foo:3', b'value:3'),
                        (b'field:foo:6', b'value:6'),
                        (b'field:foo:9', b'value:9')}

    # fetch 'field:bar:*' items expected tuple with 7 fields and 7 values
    ret = await coro(redis.ihscan(key, match=b'field:bar:*'))
    assert set(ret) == {(b'field:bar:1', b'value:1'),
                        (b'field:bar:2', b'value:2'),
                        (b'field:bar:4', b'value:4'),
                        (b'field:bar:5', b'value:5'),
                        (b'field:bar:7', b'value:7'),
                        (b'field:bar:8', b'value:8'),
                        (b'field:bar:10', b'value:10')}

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something
    ret = await coro(redis.ihscan(key, count=1))
    assert set(ret) == {(b'field:foo:3', b'value:3'),
                        (b'field:foo:6', b'value:6'),
                        (b'field:foo:9', b'value:9'),
                        (b'field:bar:1', b'value:1'),
                        (b'field:bar:2', b'value:2'),
                        (b'field:bar:4', b'value:4'),
                        (b'field:bar:5', b'value:5'),
                        (b'field:bar:7', b'value:7'),
                        (b'field:bar:8', b'value:8'),
                        (b'field:bar:10', b'value:10')}

    with pytest.raises(TypeError):
        await redis.ihscan(None)



import itertools
import pytest


@pytest.redis_version(2, 8, 0, reason='ZSCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_izscan(redis):
    key = b'key:zscan'
    scores, members = [], []

    for i in range(1, 11):
        foo_or_bar = 'bar' if i % 3 else 'foo'
        members.append('zmem:{}:{}'.format(foo_or_bar, i).encode('utf-8'))
        scores.append(i)
    pairs = list(itertools.chain(*zip(scores, members)))
    await redis.zadd(key, *pairs)
    vals = list(zip(members, scores))

    async def coro(cmd):
        lst = []
        async for i in cmd:
            lst.append(i)
        return lst

    ret = await coro(redis.izscan(key))
    assert set(ret) == set(vals)

    ret = await coro(redis.izscan(key, match=b'zmem:foo:*'))
    assert set(ret) == set(v for v in vals if b'foo' in v[0])

    ret = await coro(redis.izscan(key, match=b'zmem:bar:*'))
    assert set(ret) == set(v for v in vals if b'bar' in v[0])

    # SCAN family functions do not guarantee that the number (count) of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something

    ret = await coro(redis.izscan(key, count=2))
    assert set(ret) == set(vals)

    with pytest.raises(TypeError):
        await redis.izscan(None)



import asyncio
from asyncio import (
    StreamReaderProtocol,
    StreamReader as BaseReader,
    StreamWriter
    )

from .parser import Reader

_DEFAULT_LIMIT = 2 ** 16


@asyncio.coroutine
def open_connection(host=None, port=None, *,
                    loop=None, limit=_DEFAULT_LIMIT,
                    parser=None, **kwds):
    if loop is None:
        loop = asyncio.get_event_loop()
    reader = StreamReader(limit=limit, parser=parser, loop=loop)
    protocol = StreamReaderProtocol(reader, loop=loop)
    transport, _ = yield from loop.create_connection(
        lambda: protocol, host, port, **kwds)
    writer = StreamWriter(transport, protocol, reader, loop)
    return reader, writer


class StreamReader(BaseReader):

    def __init__(self, limit=_DEFAULT_LIMIT, parser=None, loop=None):
        super().__init__(limit, loop=loop)
        if parser is None:
            parser = Reader
        self._parser = parser()

    def feed_data(self, data):
        assert not self._eof, "feed_data after feed_eof"
        if not data:
            return

        self._parser.feed(data)
        self._wakeup_waiter()

        # TODO: implement pause_reading

    @asyncio.coroutine
    def read_response(self):
        if self._exception is not None:
            raise self._exception
        resp = self._parser.gets()
        while resp is False:
            yield from self._wait_for_data('read_response')
            resp = self._parser.gets()
        return resp



import pytest
import asyncio

from aioredis import ReplyError


@asyncio.coroutine
def push_data_with_sleep(redis, loop, key, *values):
    yield from asyncio.sleep(0.2, loop=loop)
    result = yield from redis.lpush(key, *values)
    return result


@pytest.mark.run_loop
def test_blpop(redis):
    key1, value1 = b'key:blpop:1', b'blpop:value:1'
    key2, value2 = b'key:blpop:2', b'blpop:value:2'

    # setup list
    result = yield from redis.rpush(key1, value1, value2)
    assert result == 2
    # make sure that left value poped
    test_value = yield from redis.blpop(key1)
    assert test_value == [key1, value1]
    # pop remaining value, so list should become empty
    test_value = yield from redis.blpop(key1)
    assert test_value == [key1, value2]

    with pytest.raises(TypeError):
        yield from redis.blpop(None)
    with pytest.raises(TypeError):
        yield from redis.blpop(key1, None)
    with pytest.raises(TypeError):
        yield from redis.blpop(key1, timeout=b'one')
    with pytest.raises(ValueError):
        yield from redis.blpop(key2, timeout=-10)

    # test encoding param
    yield from redis.rpush(key2, value1)
    test_value = yield from redis.blpop(key2, encoding='utf-8')
    assert test_value == ['key:blpop:2', 'blpop:value:1']


@pytest.mark.run_loop
def test_blpop_blocking_features(redis, create_redis, loop, server):
    key1, key2 = b'key:blpop:1', b'key:blpop:2'
    value = b'blpop:value:2'

    other_redis = yield from create_redis(
        server.tcp_address, loop=loop)

    # create blocking task in separate connection
    consumer = other_redis.blpop(key1, key2)

    producer_task = asyncio.Task(
        push_data_with_sleep(redis, loop, key2, value), loop=loop)
    results = yield from asyncio.gather(
        consumer, producer_task, loop=loop)

    assert results[0] == [key2, value]
    assert results[1] == 1

    # wait for data with timeout, list is emtpy, so blpop should
    # return None in 1 sec
    waiter = redis.blpop(key1, key2, timeout=1)
    test_value = yield from waiter
    assert test_value is None
    other_redis.close()


@pytest.mark.run_loop
def test_brpop(redis):
    key1, value1 = b'key:brpop:1', b'brpop:value:1'
    key2, value2 = b'key:brpop:2', b'brpop:value:2'

    # setup list
    result = yield from redis.rpush(key1, value1, value2)
    assert result == 2
    # make sure that right value poped
    test_value = yield from redis.brpop(key1)
    assert test_value == [key1, value2]
    # pop remaining value, so list should become empty
    test_value = yield from redis.brpop(key1)
    assert test_value == [key1, value1]

    with pytest.raises(TypeError):
        yield from redis.brpop(None)
    with pytest.raises(TypeError):
        yield from redis.brpop(key1, None)
    with pytest.raises(TypeError):
        yield from redis.brpop(key1, timeout=b'one')
    with pytest.raises(ValueError):
        yield from redis.brpop(key2, timeout=-10)

    # test encoding param
    yield from redis.rpush(key2, value1)
    test_value = yield from redis.brpop(key2, encoding='utf-8')
    assert test_value == ['key:brpop:2', 'brpop:value:1']


@pytest.mark.run_loop
def test_brpop_blocking_features(redis, create_redis, server, loop):
    key1, key2 = b'key:brpop:1', b'key:brpop:2'
    value = b'brpop:value:2'

    other_redis = yield from create_redis(
        server.tcp_address, loop=loop)
    # create blocking task in separate connection
    consumer_task = other_redis.brpop(key1, key2)

    producer_task = asyncio.Task(
        push_data_with_sleep(redis, loop, key2, value), loop=loop)

    results = yield from asyncio.gather(
        consumer_task, producer_task, loop=loop)

    assert results[0] == [key2, value]
    assert results[1] == 1

    # wait for data with timeout, list is emtpy, so brpop should
    # return None in 1 sec
    waiter = redis.brpop(key1, key2, timeout=1)
    test_value = yield from waiter
    assert test_value is None


@pytest.mark.run_loop
def test_brpoplpush(redis):
    key = b'key:brpoplpush:1'
    value1, value2 = b'brpoplpush:value:1', b'brpoplpush:value:2'

    destkey = b'destkey:brpoplpush:1'

    # setup list
    yield from redis.rpush(key, value1, value2)

    # move value in into head of new list
    result = yield from redis.brpoplpush(key, destkey)
    assert result == value2
    # move last value
    result = yield from redis.brpoplpush(key, destkey)
    assert result == value1

    # make sure that all values stored in new destkey list
    test_value = yield from redis.lrange(destkey, 0, -1)
    assert test_value == [value1, value2]

    with pytest.raises(TypeError):
        yield from redis.brpoplpush(None, destkey)

    with pytest.raises(TypeError):
        yield from redis.brpoplpush(key, None)

    with pytest.raises(TypeError):
        yield from redis.brpoplpush(key, destkey, timeout=b'one')

    with pytest.raises(ValueError):
        yield from redis.brpoplpush(key, destkey, timeout=-10)

    # test encoding param
    result = yield from redis.brpoplpush(
        destkey, key, encoding='utf-8')
    assert result == 'brpoplpush:value:2'


@pytest.mark.run_loop
def test_brpoplpush_blocking_features(redis, create_redis, server, loop):
    source = b'key:brpoplpush:12'
    value = b'brpoplpush:value:2'
    destkey = b'destkey:brpoplpush:2'
    other_redis = yield from create_redis(
        server.tcp_address, loop=loop)
    # create blocking task
    consumer_task = other_redis.brpoplpush(source, destkey)
    producer_task = asyncio.Task(
        push_data_with_sleep(redis, loop, source, value), loop=loop)
    results = yield from asyncio.gather(
        consumer_task, producer_task, loop=loop)
    assert results[0] == value
    assert results[1] == 1

    # make sure that all values stored in new destkey list
    test_value = yield from redis.lrange(destkey, 0, -1)
    assert test_value == [value]

    # wait for data with timeout, list is emtpy, so brpoplpush should
    # return None in 1 sec
    waiter = redis.brpoplpush(source, destkey, timeout=1)
    test_value = yield from waiter
    assert test_value is None
    other_redis.close()


@pytest.mark.run_loop
def test_lindex(redis):
    key, value = b'key:lindex:1', 'value:{}'
    # setup list
    values = [value.format(i).encode('utf-8') for i in range(0, 10)]
    yield from redis.rpush(key, *values)
    # make sure that all indexes are correct
    for i in range(0, 10):
        test_value = yield from redis.lindex(key, i)
        assert test_value == values[i]

    # get last element
    test_value = yield from redis.lindex(key, -1)
    assert test_value == b'value:9'

    # index of element if key does not exists
    test_value = yield from redis.lindex(b'not:' + key, 5)
    assert test_value is None

    # test encoding param
    yield from redis.rpush(key, 'one', 'two')
    test_value = yield from redis.lindex(key, 10, encoding='utf-8')
    assert test_value == 'one'
    test_value = yield from redis.lindex(key, 11, encoding='utf-8')
    assert test_value == 'two'

    with pytest.raises(TypeError):
        yield from redis.lindex(None, -1)

    with pytest.raises(TypeError):
        yield from redis.lindex(key, b'one')


@pytest.mark.run_loop
def test_linsert(redis):
    key = b'key:linsert:1'
    value1, value2, value3, value4 = b'Hello', b'World', b'foo', b'bar'
    yield from redis.rpush(key, value1, value2)

    # insert element before pivot
    test_value = yield from redis.linsert(
        key, value2, value3, before=True)
    assert test_value == 3
    # insert element after pivot
    test_value = yield from redis.linsert(
        key, value2, value4, before=False)
    assert test_value == 4

    # make sure that values actually inserted in right placed
    test_value = yield from redis.lrange(key, 0, -1)
    expected = [value1, value3, value2, value4]
    assert test_value == expected

    # try to insert something when pivot value does not exits
    test_value = yield from redis.linsert(
        key, b'not:pivot', value3, before=True)
    assert test_value == -1

    with pytest.raises(TypeError):
        yield from redis.linsert(None, value1, value3)


@pytest.mark.run_loop
def test_llen(redis):
    key = b'key:llen:1'
    value1, value2 = b'Hello', b'World'
    yield from redis.rpush(key, value1, value2)

    test_value = yield from redis.llen(key)
    assert test_value == 2

    test_value = yield from redis.llen(b'not:' + key)
    assert test_value == 0

    with pytest.raises(TypeError):
        yield from redis.llen(None)


@pytest.mark.run_loop
def test_lpop(redis):
    key = b'key:lpop:1'
    value1, value2 = b'lpop:value:1', b'lpop:value:2'

    # setup list
    result = yield from redis.rpush(key, value1, value2)
    assert result == 2
    # make sure that left value poped
    test_value = yield from redis.lpop(key)
    assert test_value == value1
    # pop remaining value, so list should become empty
    test_value = yield from redis.lpop(key)
    assert test_value == value2
    # pop from empty list
    test_value = yield from redis.lpop(key)
    assert test_value is None

    # test encoding param
    yield from redis.rpush(key, 'value')
    test_value = yield from redis.lpop(key, encoding='utf-8')
    assert test_value == 'value'

    with pytest.raises(TypeError):
        yield from redis.lpop(None)


@pytest.mark.run_loop
def test_lpush(redis):
    key = b'key:lpush'
    value1, value2 = b'value:1', b'value:2'

    # add multiple values to the list, with key that does not exists
    result = yield from redis.lpush(key, value1, value2)
    assert result == 2

    # make sure that values actually inserted in right placed and order
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [value2, value1]

    # test encoding param
    test_value = yield from redis.lrange(key, 0, -1, encoding='utf-8')
    assert test_value == ['value:2', 'value:1']

    with pytest.raises(TypeError):
        yield from redis.lpush(None, value1)


@pytest.mark.run_loop
def test_lpushx(redis):
    key = b'key:lpushx'
    value1, value2 = b'value:1', b'value:2'

    # add multiple values to the list, with key that does not exists
    # so value should not be pushed
    result = yield from redis.lpushx(key, value2)
    assert result == 0
    # init key with list by using regular lpush
    result = yield from redis.lpush(key, value1)
    assert result == 1

    result = yield from redis.lpushx(key, value2)
    assert result == 2

    # make sure that values actually inserted in right placed and order
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [value2, value1]

    with pytest.raises(TypeError):
        yield from redis.lpushx(None, value1)


@pytest.mark.run_loop
def test_lrange(redis):
    key, value = b'key:lrange:1', 'value:{}'
    values = [value.format(i).encode('utf-8') for i in range(0, 10)]
    yield from redis.rpush(key, *values)

    test_value = yield from redis.lrange(key, 0, 2)
    assert test_value == values[0:3]

    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == values

    test_value = yield from redis.lrange(key, -2, -1)
    assert test_value == values[-2:]

    # range of elements if key does not exists
    test_value = yield from redis.lrange(b'not:' + key, 0, -1)
    assert test_value == []

    with pytest.raises(TypeError):
        yield from redis.lrange(None, 0, -1)

    with pytest.raises(TypeError):
        yield from redis.lrange(key, b'zero', -1)

    with pytest.raises(TypeError):
        yield from redis.lrange(key, 0, b'one')


@pytest.mark.run_loop
def test_lrem(redis):
    key, value = b'key:lrem:1', 'value:{}'
    values = [value.format(i % 2).encode('utf-8') for i in range(0, 10)]
    yield from redis.rpush(key, *values)
    # remove elements from tail to head
    test_value = yield from redis.lrem(key, -4, b'value:0')
    assert test_value == 4
    # remove element from head to tail
    test_value = yield from redis.lrem(key, 4, b'value:1')
    assert test_value == 4

    # remove values that not in list
    test_value = yield from redis.lrem(key, 4, b'value:other')
    assert test_value == 0

    # make sure that only two values left in the list
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [b'value:0', b'value:1']

    # remove all instance of value:0
    test_value = yield from redis.lrem(key, 0, b'value:0')
    assert test_value == 1

    # make sure that only one values left in the list
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [b'value:1']

    with pytest.raises(TypeError):
        yield from redis.lrem(None, 0, b'value:0')

    with pytest.raises(TypeError):
        yield from redis.lrem(key, b'ten', b'value:0')


@pytest.mark.run_loop
def test_lset(redis):
    key, value = b'key:lset', 'value:{}'
    values = [value.format(i).encode('utf-8') for i in range(0, 3)]
    yield from redis.rpush(key, *values)

    yield from redis.lset(key, 0, b'foo')
    yield from redis.lset(key, -1, b'baz')
    yield from redis.lset(key, -2, b'zap')

    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [b'foo', b'zap', b'baz']

    with pytest.raises(TypeError):
        yield from redis.lset(None, 0, b'value:0')

    with pytest.raises(ReplyError):
        yield from redis.lset(key, 100, b'value:0')

    with pytest.raises(TypeError):
        yield from redis.lset(key, b'one', b'value:0')


@pytest.mark.run_loop
def test_ltrim(redis):
    key, value = b'key:ltrim', 'value:{}'
    values = [value.format(i).encode('utf-8') for i in range(0, 10)]
    yield from redis.rpush(key, *values)

    # trim with negative indexes
    yield from redis.ltrim(key, 0, -5)
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == values[:-4]
    # trim with positive indexes
    yield from redis.ltrim(key, 0, 2)
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == values[:3]

    # try to trim out of range indexes
    res = yield from redis.ltrim(key, 100, 110)
    assert res is True
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == []

    with pytest.raises(TypeError):
        yield from redis.ltrim(None, 0, -1)

    with pytest.raises(TypeError):
        yield from redis.ltrim(key, b'zero', -1)

    with pytest.raises(TypeError):
        yield from redis.ltrim(key, 0, b'one')


@pytest.mark.run_loop
def test_rpop(redis):
    key = b'key:rpop:1'
    value1, value2 = b'rpop:value:1', b'rpop:value:2'

    # setup list
    result = yield from redis.rpush(key, value1, value2)
    assert result == 2
    # make sure that left value poped
    test_value = yield from redis.rpop(key)
    assert test_value == value2
    # pop remaining value, so list should become empty
    test_value = yield from redis.rpop(key)
    assert test_value == value1
    # pop from empty list
    test_value = yield from redis.rpop(key)
    assert test_value is None

    # test encoding param
    yield from redis.rpush(key, 'value')
    test_value = yield from redis.rpop(key, encoding='utf-8')
    assert test_value == 'value'

    with pytest.raises(TypeError):
        yield from redis.rpop(None)


@pytest.mark.run_loop
def test_rpoplpush(redis):
    key = b'key:rpoplpush:1'
    value1, value2 = b'rpoplpush:value:1', b'rpoplpush:value:2'
    destkey = b'destkey:rpoplpush:1'

    # setup list
    yield from redis.rpush(key, value1, value2)

    # move value in into head of new list
    result = yield from redis.rpoplpush(key, destkey)
    assert result == value2
    # move last value
    result = yield from redis.rpoplpush(key, destkey)
    assert result == value1

    # make sure that all values stored in new destkey list
    result = yield from redis.lrange(destkey, 0, -1)
    assert result == [value1, value2]

    # test encoding param
    result = yield from redis.rpoplpush(
        destkey, key, encoding='utf-8')
    assert result == 'rpoplpush:value:2'

    with pytest.raises(TypeError):
        yield from redis.rpoplpush(None, destkey)

    with pytest.raises(TypeError):
        yield from redis.rpoplpush(key, None)


@pytest.mark.run_loop
def test_rpush(redis):
    key = b'key:rpush'
    value1, value2 = b'value:1', b'value:2'

    # add multiple values to the list, with key that does not exists
    result = yield from redis.rpush(key, value1, value2)
    assert result == 2

    # make sure that values actually inserted in right placed and order
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [value1, value2]

    with pytest.raises(TypeError):
        yield from redis.rpush(None, value1)


@pytest.mark.run_loop
def test_rpushx(redis):
    key = b'key:rpushx'
    value1, value2 = b'value:1', b'value:2'

    # add multiple values to the list, with key that does not exists
    # so value should not be pushed
    result = yield from redis.rpushx(key, value2)
    assert result == 0
    # init key with list by using regular rpush
    result = yield from redis.rpush(key, value1)
    assert result == 1

    result = yield from redis.rpushx(key, value2)
    assert result == 2

    # make sure that values actually inserted in right placed and order
    test_value = yield from redis.lrange(key, 0, -1)
    assert test_value == [value1, value2]

    with pytest.raises(TypeError):
        yield from redis.rpushx(None, value1)



from aioredis.util import wait_convert, PY_35

if PY_35:
    from aioredis.util import _ScanIterPairs


class SortedSetCommandsMixin:
    """Sorted Sets commands mixin.

    For commands details see: http://redis.io/commands/#sorted_set
    """

    ZSET_EXCLUDE_MIN = 'ZSET_EXCLUDE_MIN'
    ZSET_EXCLUDE_MAX = 'ZSET_EXCLUDE_MAX'
    ZSET_EXCLUDE_BOTH = 'ZSET_EXCLUDE_BOTH'

    ZSET_AGGREGATE_SUM = 'ZSET_AGGREGATE_SUM'
    ZSET_AGGREGATE_MIN = 'ZSET_AGGREGATE_MIN'
    ZSET_AGGREGATE_MAX = 'ZSET_AGGREGATE_MAX'

    def zadd(self, key, score, member, *pairs):
        """Add one or more members to a sorted set or update its score.

        :raises TypeError: score not int or float
        :raises TypeError: length of pairs is not even number
        """
        if not isinstance(score, (int, float)):
            raise TypeError("score argument must be int or float")
        if len(pairs) % 2 != 0:
            raise TypeError("length of pairs must be even number")

        scores = (item for i, item in enumerate(pairs) if i % 2 == 0)
        if any(not isinstance(s, (int, float)) for s in scores):
            raise TypeError("all scores must be int or float")
        return self.execute(b'ZADD', key, score, member, *pairs)

    def zcard(self, key):
        """Get the number of members in a sorted set."""
        return self.execute(b'ZCARD', key)

    def zcount(self, key, min=float('-inf'), max=float('inf'),
               *, exclude=None):
        """Count the members in a sorted set with scores
        within the given values.

        :raises TypeError: min or max is not float or int
        :raises ValueError: if min grater then max
        """
        if not isinstance(min, (int, float)):
            raise TypeError("min argument must be int or float")
        if not isinstance(max, (int, float)):
            raise TypeError("max argument must be int or float")
        if min > max:
            raise ValueError("min could not be grater then max")
        return self.execute(b'ZCOUNT', key,
                            *_encode_min_max(exclude, min, max))

    def zincrby(self, key, increment, member):
        """Increment the score of a member in a sorted set.

        :raises TypeError: increment is not float or int
        """
        if not isinstance(increment, (int, float)):
            raise TypeError("increment argument must be int or float")
        fut = self.execute(b'ZINCRBY', key, increment, member)
        return wait_convert(fut, int_or_float)

    def zinterstore(self, destkey, key, *keys,
                    with_weights=False, aggregate=None):
        """Intersect multiple sorted sets and store result in a new key.

        :param bool with_weights: when set to true each key must be a tuple
                                  in form of (key, weight)
        """
        keys = (key,) + keys
        numkeys = len(keys)
        args = []
        if with_weights:
            assert all(isinstance(val, (list, tuple)) for val in keys), (
                "All key arguments must be (key, weight) tuples")
            weights = ['WEIGHTS']
            for key, weight in keys:
                args.append(key)
                weights.append(weight)
            args.extend(weights)
        else:
            args.extend(keys)

        if aggregate is self.ZSET_AGGREGATE_SUM:
            args.extend(('AGGREGATE', 'SUM'))
        elif aggregate is self.ZSET_AGGREGATE_MAX:
            args.extend(('AGGREGATE', 'MAX'))
        elif aggregate is self.ZSET_AGGREGATE_MIN:
            args.extend(('AGGREGATE', 'MIN'))
        fut = self.execute(b'ZINTERSTORE', destkey, numkeys, *args)
        return fut

    def zlexcount(self, key, min=b'-', max=b'+', include_min=True,
                  include_max=True):
        """Count the number of members in a sorted set between a given
        lexicographical range.

        :raises TypeError: if min is not bytes
        :raises TypeError: if max is not bytes
        """
        if not isinstance(min, bytes):  # FIXME
            raise TypeError("min argument must be bytes")
        if not isinstance(max, bytes):  # FIXME     Why only bytes?
            raise TypeError("max argument must be bytes")
        if not min == b'-':
            min = (b'[' if include_min else b'(') + min
        if not max == b'+':
            max = (b'[' if include_max else b'(') + max
        return self.execute(b'ZLEXCOUNT', key, min, max)

    def zrange(self, key, start=0, stop=-1, withscores=False):
        """Return a range of members in a sorted set, by index.

        :raises TypeError: if start is not int
        :raises TypeError: if stop is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(stop, int):
            raise TypeError("stop argument must be int")
        if withscores:
            args = [b'WITHSCORES']
        else:
            args = []
        fut = self.execute(b'ZRANGE', key, start, stop, *args)
        if withscores:
            return wait_convert(fut, pairs_int_or_float)
        return fut

    def zrangebylex(self, key, min=b'-', max=b'+', include_min=True,
                    include_max=True, offset=None, count=None):
        """Return a range of members in a sorted set, by lexicographical range.

        :raises TypeError: if min is not bytes
        :raises TypeError: if max is not bytes
        :raises TypeError: if both offset and count are not specified
        :raises TypeError: if offset is not bytes
        :raises TypeError: if count is not bytes
        """
        if not isinstance(min, bytes):  # FIXME
            raise TypeError("min argument must be bytes")
        if not isinstance(max, bytes):  # FIXME
            raise TypeError("max argument must be bytes")
        if not min == b'-':
            min = (b'[' if include_min else b'(') + min
        if not max == b'+':
            max = (b'[' if include_max else b'(') + max

        if (offset is not None and count is None) or \
                (count is not None and offset is None):
            raise TypeError("offset and count must both be specified")
        if offset is not None and not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if count is not None and not isinstance(count, int):
            raise TypeError("count argument must be int")

        args = []
        if offset is not None and count is not None:
            args.extend([b'LIMIT', offset, count])

        return self.execute(b'ZRANGEBYLEX', key, min, max, *args)

    def zrangebyscore(self, key, min=float('-inf'), max=float('inf'),
                      withscores=False, offset=None, count=None,
                      *, exclude=None):
        """Return a range of members in a sorted set, by score.

        :raises TypeError: if min or max is not float or int
        :raises TypeError: if both offset and count are not specified
        :raises TypeError: if offset is not int
        :raises TypeError: if count is not int
        """
        if not isinstance(min, (int, float)):
            raise TypeError("min argument must be int or float")
        if not isinstance(max, (int, float)):
            raise TypeError("max argument must be int or float")

        if (offset is not None and count is None) or \
                (count is not None and offset is None):
            raise TypeError("offset and count must both be specified")
        if offset is not None and not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if count is not None and not isinstance(count, int):
            raise TypeError("count argument must be int")

        min, max = _encode_min_max(exclude, min, max)

        args = []
        if withscores:
            args = [b'WITHSCORES']
        if offset is not None and count is not None:
            args.extend([b'LIMIT', offset, count])
        fut = self.execute(b'ZRANGEBYSCORE', key, min, max, *args)
        if withscores:
            return wait_convert(fut, pairs_int_or_float)
        return fut

    def zrank(self, key, member):
        """Determine the index of a member in a sorted set."""
        return self.execute(b'ZRANK', key, member)

    def zrem(self, key, member, *members):
        """Remove one or more members from a sorted set."""
        return self.execute(b'ZREM', key, member, *members)

    def zremrangebylex(self, key, min=b'-', max=b'+',
                       include_min=True, include_max=True):
        """Remove all members in a sorted set between the given
        lexicographical range.

        :raises TypeError: if min is not bytes
        :raises TypeError: if max is not bytes
        """
        if not isinstance(min, bytes):  # FIXME
            raise TypeError("min argument must be bytes")
        if not isinstance(max, bytes):  # FIXME
            raise TypeError("max argument must be bytes")
        if not min == b'-':
            min = (b'[' if include_min else b'(') + min
        if not max == b'+':
            max = (b'[' if include_max else b'(') + max
        return self.execute(b'ZREMRANGEBYLEX', key, min, max)

    def zremrangebyrank(self, key, start, stop):
        """Remove all members in a sorted set within the given indexes.

        :raises TypeError: if start is not int
        :raises TypeError: if stop is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(stop, int):
            raise TypeError("stop argument must be int")
        return self.execute(b'ZREMRANGEBYRANK', key, start, stop)

    def zremrangebyscore(self, key, min=float('-inf'), max=float('inf'),
                         *, exclude=None):
        """Remove all members in a sorted set within the given scores.

        :raises TypeError: if min or max is not int or float
        """
        if not isinstance(min, (int, float)):
            raise TypeError("min argument must be int or float")
        if not isinstance(max, (int, float)):
            raise TypeError("max argument must be int or float")

        min, max = _encode_min_max(exclude, min, max)
        return self.execute(b'ZREMRANGEBYSCORE', key, min, max)

    def zrevrange(self, key, start, stop, withscores=False):
        """Return a range of members in a sorted set, by index,
        with scores ordered from high to low.

        :raises TypeError: if start or stop is not int
        """
        if not isinstance(start, int):
            raise TypeError("start argument must be int")
        if not isinstance(stop, int):
            raise TypeError("stop argument must be int")
        if withscores:
            args = [b'WITHSCORES']
        else:
            args = []
        fut = self.execute(b'ZREVRANGE', key, start, stop, *args)
        if withscores:
            return wait_convert(fut, pairs_int_or_float)
        return fut

    def zrevrangebyscore(self, key, max=float('inf'), min=float('-inf'),
                         *, exclude=None, withscores=False,
                         offset=None, count=None):
        """Return a range of members in a sorted set, by score,
        with scores ordered from high to low.

        :raises TypeError: if min or max is not float or int
        :raises TypeError: if both offset and count are not specified
        :raises TypeError: if offset is not int
        :raises TypeError: if count is not int
        """
        if not isinstance(min, (int, float)):
            raise TypeError("min argument must be int or float")
        if not isinstance(max, (int, float)):
            raise TypeError("max argument must be int or float")

        if (offset is not None and count is None) or \
                (count is not None and offset is None):
            raise TypeError("offset and count must both be specified")
        if offset is not None and not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if count is not None and not isinstance(count, int):
            raise TypeError("count argument must be int")

        min, max = _encode_min_max(exclude, min, max)

        args = []
        if withscores:
            args = [b'WITHSCORES']
        if offset is not None and count is not None:
            args.extend([b'LIMIT', offset, count])
        fut = self.execute(b'ZREVRANGEBYSCORE', key, max, min, *args)
        if withscores:
            return wait_convert(fut, pairs_int_or_float)
        return fut

    def zrevrangebylex(self, key, min=b'-', max=b'+', include_min=True,
                       include_max=True, offset=None, count=None):
        """Return a range of members in a sorted set, by lexicographical range
        from high to low.

        :raises TypeError: if min is not bytes
        :raises TypeError: if max is not bytes
        :raises TypeError: if both offset and count are not specified
        :raises TypeError: if offset is not bytes
        :raises TypeError: if count is not bytes
        """
        if not isinstance(min, bytes):  # FIXME
            raise TypeError("min argument must be bytes")
        if not isinstance(max, bytes):  # FIXME
            raise TypeError("max argument must be bytes")
        if not min == b'-':
            min = (b'[' if include_min else b'(') + min
        if not max == b'+':
            max = (b'[' if include_max else b'(') + max

        if (offset is not None and count is None) or \
                (count is not None and offset is None):
            raise TypeError("offset and count must both be specified")
        if offset is not None and not isinstance(offset, int):
            raise TypeError("offset argument must be int")
        if count is not None and not isinstance(count, int):
            raise TypeError("count argument must be int")

        args = []
        if offset is not None and count is not None:
            args.extend([b'LIMIT', offset, count])

        return self.execute(b'ZREVRANGEBYLEX', key, max, min, *args)

    def zrevrank(self, key, member):
        """Determine the index of a member in a sorted set, with
        scores ordered from high to low.
        """
        return self.execute(b'ZREVRANK', key, member)

    def zscore(self, key, member):
        """Get the score associated with the given member in a sorted set."""
        fut = self.execute(b'ZSCORE', key, member)
        return wait_convert(fut, optional_int_or_float)

    def zunionstore(self, destkey, key, *keys,
                    with_weights=False, aggregate=None):
        """Add multiple sorted sets and store result in a new key."""
        keys = (key,) + keys
        numkeys = len(keys)
        args = []
        if with_weights:
            assert all(isinstance(val, (list, tuple)) for val in keys), (
                "All key arguments must be (key, weight) tuples")
            weights = ['WEIGHTS']
            for key, weight in keys:
                args.append(key)
                weights.append(weight)
            args.extend(weights)
        else:
            args.extend(keys)

        if aggregate is self.ZSET_AGGREGATE_SUM:
            args.extend(('AGGREGATE', 'SUM'))
        elif aggregate is self.ZSET_AGGREGATE_MAX:
            args.extend(('AGGREGATE', 'MAX'))
        elif aggregate is self.ZSET_AGGREGATE_MIN:
            args.extend(('AGGREGATE', 'MIN'))
        fut = self.execute(b'ZUNIONSTORE', destkey, numkeys, *args)
        return fut

    def zscan(self, key, cursor=0, match=None, count=None):
        """Incrementally iterate sorted sets elements and associated scores."""
        args = []
        if match is not None:
            args += [b'MATCH', match]
        if count is not None:
            args += [b'COUNT', count]
        fut = self.execute(b'ZSCAN', key, cursor, *args)

        def _converter(obj):
            return (int(obj[0]), pairs_int_or_float(obj[1]))

        return wait_convert(fut, _converter)

    if PY_35:
        def izscan(self, key, *, match=None, count=None):
            """Incrementally iterate sorted set items using async for.

            Usage example:

            >>> async for val, score in redis.izscan(key, match='something*'):
            ...     print('Matched:', val, ':', score)

            """
            return _ScanIterPairs(lambda cur: self.zscan(key, cur,
                                                         match=match,
                                                         count=count))


def _encode_min_max(flag, min, max):
    if flag is SortedSetCommandsMixin.ZSET_EXCLUDE_MIN:
        return '({}'.format(min), max
    elif flag is SortedSetCommandsMixin.ZSET_EXCLUDE_MAX:
        return min, '({}'.format(max)
    elif flag is SortedSetCommandsMixin.ZSET_EXCLUDE_BOTH:
        return '({}'.format(min), '({}'.format(max)
    return min, max


def int_or_float(value):
    assert isinstance(value, (str, bytes)), 'raw_value must be bytes'
    try:
        return int(value)
    except ValueError:
        return float(value)


def optional_int_or_float(value):
    if value is None:
        return value
    return int_or_float(value)


def pairs_int_or_float(value):
    it = iter(value)
    return list(sum(([val, int_or_float(score)] for val, score in zip(it, it)),
                    []))



import os
import sys
import logging


logger = logging.getLogger('aioredis')
sentinel_logger = logger.getChild('sentinel')

if os.environ.get("AIOREDIS_DEBUG"):
    logger.setLevel(logging.DEBUG)
    handler = logging.StreamHandler(stream=sys.stderr)
    handler.setFormatter(logging.Formatter(
        "%(asctime)s %(name)s %(levelname)s %(message)s"))
    logger.addHandler(handler)
    os.environ["AIOREDIS_DEBUG"] = ""



import pytest

from aioredis.util import coerced_keys_dict


def test_simple():
    d = coerced_keys_dict()
    assert d == {}

    d = coerced_keys_dict({b'a': 'b', b'c': 'd'})
    assert 'a' in d
    assert b'a' in d
    assert 'c' in d
    assert b'c' in d
    assert d == {b'a': 'b', b'c': 'd'}


def test_invalid_init():
    d = coerced_keys_dict({'foo': 'bar'})
    assert d == {'foo': 'bar'}

    assert 'foo' not in d
    assert b'foo' not in d
    with pytest.raises(KeyError):
        d['foo']
    with pytest.raises(KeyError):
        d[b'foo']

    d = coerced_keys_dict()
    d.update({'foo': 'bar'})
    assert d == {'foo': 'bar'}

    assert 'foo' not in d
    assert b'foo' not in d
    with pytest.raises(KeyError):
        d['foo']
    with pytest.raises(KeyError):
        d[b'foo']


def test_valid_init():
    d = coerced_keys_dict({b'foo': 'bar'})
    assert d == {b'foo': 'bar'}
    assert 'foo' in d
    assert b'foo' in d
    assert d['foo'] == 'bar'
    assert d[b'foo'] == 'bar'

    d = coerced_keys_dict()
    d.update({b'foo': 'bar'})
    assert d == {b'foo': 'bar'}
    assert 'foo' in d
    assert b'foo' in d
    assert d['foo'] == 'bar'
    assert d[b'foo'] == 'bar'



import asyncio
import contextlib

from concurrent.futures import ALL_COMPLETED
from async_timeout import timeout as async_timeout

from ..log import sentinel_logger
from ..util import async_task
from ..pubsub import Receiver
from ..pool import create_pool, ConnectionsPool
from ..errors import MasterNotFoundError, SlaveNotFoundError, PoolClosedError


# Address marker for discovery
_NON_DISCOVERED = object()

_logger = sentinel_logger.getChild('monitor')


@asyncio.coroutine
def create_sentinel_pool(sentinels, *, db=None, password=None,
                         encoding=None, minsize=1, maxsize=10,
                         ssl=None, parser=None, timeout=0.2, loop=None):
    """Create SentinelPool."""
    # FIXME: revise default timeout value
    assert isinstance(sentinels, (list, tuple)), sentinels
    if loop is None:
        loop = asyncio.get_event_loop()

    pool = SentinelPool(sentinels, db=db,
                        password=password,
                        ssl=ssl,
                        encoding=encoding,
                        parser=parser,
                        minsize=minsize,
                        maxsize=maxsize,
                        timeout=timeout,
                        loop=loop)
    yield from pool.discover()
    return pool


class SentinelPool:
    """Sentinel connections pool.

    Holds connection pools to known and discovered (TBD) Sentinels
    as well as services' connections.
    """

    def __init__(self, sentinels, *, db=None, password=None, ssl=None,
                 encoding=None, parser=None, minsize, maxsize, timeout,
                 loop=None):
        if loop is None:
            loop = asyncio.get_event_loop()
        # TODO: add connection/discover timeouts;
        #       and what to do if no master is found:
        #       (raise error or try forever or try until timeout)

        # XXX: _sentinels is unordered
        self._sentinels = set(sentinels)
        self._loop = loop
        self._timeout = timeout
        self._pools = []     # list of sentinel pools
        self._masters = {}
        self._slaves = {}
        self._parser_class = parser
        self._redis_db = db
        self._redis_password = password
        self._redis_ssl = ssl
        self._redis_encoding = encoding
        self._redis_minsize = minsize
        self._redis_maxsize = maxsize
        self._close_state = asyncio.Event(loop=loop)
        self._close_waiter = None
        self._monitor = monitor = Receiver(loop=loop)

        @asyncio.coroutine
        def echo_events():
            try:
                while (yield from monitor.wait_message()):
                    ch, (ev, data) = yield from monitor.get(encoding='utf-8')
                    ev = ev.decode('utf-8')
                    _logger.debug("%s: %s", ev, data)
                    if ev in ('+odown',):
                        typ, name, *tail = data.split(' ')
                        if typ == 'master':
                            self._need_rediscover(name)
                # TODO: parse messages;
                #   watch +new-epoch which signals `failover in progres`
                #   freeze reconnection
                #   wait / discover new master (find proper way)
                #   unfreeze reconnection
                #
                #   discover master in default way
                #       get-master-addr...
                #       connnect
                #       role
                #       etc...
            except asyncio.CancelledError:
                pass
        self._monitor_task = async_task(echo_events(), loop=loop)

    @property
    def discover_timeout(self):
        """Timeout (seconds) for Redis/Sentinel command calls during
        master/slave address discovery.
        """
        return self._timeout

    def master_for(self, service):
        """Returns wrapper to master's pool for requested service."""
        # TODO: make it coroutine and connect minsize connections
        if service not in self._masters:
            self._masters[service] = ManagedPool(
                self, service, is_master=True,
                db=self._redis_db,
                password=self._redis_password,
                encoding=self._redis_encoding,
                minsize=self._redis_minsize,
                maxsize=self._redis_maxsize,
                ssl=self._redis_ssl,
                parser=self._parser_class,
                loop=self._loop)
        return self._masters[service]

    def slave_for(self, service):
        """Returns wrapper to slave's pool for requested service."""
        # TODO: make it coroutine and connect minsize connections
        if service not in self._slaves:
            self._slaves[service] = ManagedPool(
                self, service, is_master=False,
                db=self._redis_db,
                password=self._redis_password,
                encoding=self._redis_encoding,
                minsize=self._redis_minsize,
                maxsize=self._redis_maxsize,
                ssl=self._redis_ssl,
                parser=self._parser_class,
                loop=self._loop)
        return self._slaves[service]

    def execute(self, command, *args, **kwargs):
        """Execute sentinel command."""
        # TODO: choose pool
        #   kwargs can be used to control which sentinel to use
        if self.closed:
            raise PoolClosedError("Sentinel pool is closed")
        for pool in self._pools:
            return pool.execute(command, *args, **kwargs)
        # how to handle errors and pick other pool?
        #   is the only way to make it coroutine?

    @property
    def closed(self):
        """True if pool is closed or closing."""
        return self._close_state.is_set()

    def close(self):
        """Close all controlled connections (both sentinel and redis)."""
        if not self._close_state.is_set():
            self._close_waiter = async_task(self._do_close(), loop=self._loop)
            self._close_state.set()

    @asyncio.coroutine
    def _do_close(self):
        yield from self._close_state.wait()
        # TODO: lock
        tasks = []
        task, self._monitor_task = self._monitor_task, None
        task.cancel()
        tasks.append(task)
        while self._pools:
            pool = self._pools.pop(0)
            pool.close()
            tasks.append(pool.wait_closed())
        while self._masters:
            _, pool = self._masters.popitem()
            pool.close()
            tasks.append(pool.wait_closed())
        while self._slaves:
            _, pool = self._slaves.popitem()
            pool.close()
            tasks.append(pool.wait_closed())
        yield from asyncio.gather(*tasks, loop=self._loop)

    @asyncio.coroutine
    def wait_closed(self):
        """Wait until pool gets closed."""
        yield from self._close_state.wait()
        assert self._close_waiter is not None
        yield from asyncio.shield(self._close_waiter, loop=self._loop)

    @asyncio.coroutine
    def discover(self, timeout=0.2):    # TODO: better name?
        """Discover sentinels and all monitored services within given timeout.

        If no sentinels discovered within timeout: TimeoutError is raised.
        If some sentinels were discovered but not all — it is ok.
        If not all monitored services (masters/slaves) discovered
        (or connections established) — it is ok.
        TBD: what if some sentinels/services unreachable;
        """
        # TODO: check not closed
        # TODO: discovery must be done with some customizable timeout.
        tasks = []
        pools = []
        for addr in self._sentinels:    # iterate over unordered set
            tasks.append(self._connect_sentinel(addr, timeout, pools))
        done, pending = yield from asyncio.wait(tasks, loop=self._loop,
                                                return_when=ALL_COMPLETED)
        assert not pending, ("Expected all tasks to complete", done, pending)

        for task in done:
            result = task.result()
            if isinstance(result, Exception):
                continue    # FIXME
        if not pools:
            raise Exception("Could not connect to any sentinel")
        pools, self._pools[:] = self._pools[:], pools
        # TODO: close current connections
        for pool in pools:
            pool.close()
            yield from pool.wait_closed()

        # TODO: discover peer sentinels
        for pool in self._pools:
            yield from pool.execute_pubsub(
                b'psubscribe', self._monitor.pattern('*'))

    @asyncio.coroutine
    def _connect_sentinel(self, address, timeout, pools):
        """Try to connect to specified Sentinel returning either
        connections pool or exception.
        """
        try:
            with async_timeout(timeout, loop=self._loop):
                pool = yield from create_pool(
                    address, minsize=1, maxsize=2,
                    parser=self._parser_class,
                    loop=self._loop)
            pools.append(pool)
            return pool
        except asyncio.TimeoutError as err:
            sentinel_logger.debug(
                "Failed to connect to Sentinel(%r) within %ss timeout",
                address, timeout)
            return err
        except Exception as err:
            sentinel_logger.debug(
                "Error connecting to Sentinel(%r): %r", address, err)
            return err

    @asyncio.coroutine
    def discover_master(self, service, timeout):
        """Perform Master discovery for specified service."""
        # TODO: get lock
        idle_timeout = timeout
        # FIXME: single timeout used 4 times;
        #   meaning discovery can take up to:
        #   3 * timeout * (sentinels count)
        #
        #   having one global timeout also can leed to
        #   a problem when not all sentinels are checked.

        # use a copy, cause pools can change
        pools = self._pools[:]
        for sentinel in pools:
            try:
                with async_timeout(timeout, loop=self._loop):
                    address = yield from self._get_masters_address(
                        sentinel, service)

                pool = self._masters[service]
                with async_timeout(timeout, loop=self._loop), \
                        contextlib.ExitStack() as stack:
                    conn = yield from pool._create_new_connection(address)
                    stack.callback(conn.close)
                    yield from self._verify_service_role(conn, 'master')
                    stack.pop_all()

                return conn
            except asyncio.CancelledError:
                # we must correctly handle CancelledError(s):
                #   application may be stopped or function can be cancelled
                #   by outer timeout, so we must stop the look up.
                raise
            except asyncio.TimeoutError:
                continue
            except DiscoverError as err:
                sentinel_logger.debug("DiscoverError(%r, %s): %r",
                                      sentinel, service, err)
                yield from asyncio.sleep(idle_timeout, loop=self._loop)
                continue
            except Exception:
                # TODO: clear (drop) connections to schedule reconnect
                yield from asyncio.sleep(idle_timeout, loop=self._loop)
                continue
        else:
            raise MasterNotFoundError("No master found for {}".format(service))

    @asyncio.coroutine
    def discover_slave(self, service, timeout, **kwargs):
        """Perform Slave discovery for specified service."""
        # TODO: use kwargs to change how slaves are picked up
        #   (eg: round-robin, priority, random, etc)
        idle_timeout = timeout
        pools = self._pools[:]
        for sentinel in pools:
            try:
                with async_timeout(timeout, loop=self._loop):
                    address = yield from self._get_slave_address(
                        sentinel, service)  # add **kwargs
                pool = self._slaves[service]
                with async_timeout(timeout, loop=self._loop), \
                        contextlib.ExitStack() as stack:
                    conn = yield from pool._create_new_connection(address)
                    stack.callback(conn.close)
                    yield from self._verify_service_role(conn, 'slave')
                    stack.pop_all()
                return conn
            except asyncio.CancelledError:
                raise
            except asyncio.TimeoutError:
                continue
            except DiscoverError:
                yield from asyncio.sleep(idle_timeout, loop=self._loop)
                continue
            except Exception:
                yield from asyncio.sleep(idle_timeout, loop=self._loop)
                continue
        raise SlaveNotFoundError("No slave found for {}".format(service))

    @asyncio.coroutine
    def _get_masters_address(self, sentinel, service):
        # NOTE: we don't use `get-master-addr-by-name`
        #   as it can provide stale data so we repeat
        #   after redis-py and check service flags.
        state = yield from sentinel.execute(b'sentinel', b'master',
                                            service, encoding='utf-8')
        if not state:
            raise UnknownService()
        state = make_dict(state)
        address = state['ip'], int(state['port'])
        flags = set(state['flags'].split(','))
        if {'s_down', 'o_down', 'disconnected'} & flags:
            raise BadState(state)
        return address

    @asyncio.coroutine
    def _get_slave_address(self, sentinel, service):
        # Find and return single slave address
        slaves = yield from sentinel.execute(b'sentinel', b'slaves',
                                             service, encoding='utf-8')
        if not slaves:
            raise UnknownService()
        for state in map(make_dict, slaves):
            address = state['ip'], int(state['port'])
            flags = set(state['flags'].split(','))
            if {'s_down', 'o_down', 'disconnected'} & flags:
                continue
            return address
        else:
            raise BadState(state)   # XXX: only last state

    @asyncio.coroutine
    def _verify_service_role(self, conn, role):
        res = yield from conn.execute(b'role', encoding='utf-8')
        if res[0] != role:
            raise RoleMismatch(res)

    def _need_rediscover(self, service):
        sentinel_logger.debug("Must redisover service %s", service)
        for service, pool in self._masters.items():
            pool.need_rediscover()
        for service, pool in self._slaves.items():
            pool.need_rediscover()


class ManagedPool(ConnectionsPool):

    def __init__(self, sentinel, service, is_master,
                 db=None, password=None, encoding=None, parser=None,
                 *, minsize, maxsize, ssl=None, loop=None):
        super().__init__(_NON_DISCOVERED,
                         db=db, password=password, encoding=encoding,
                         minsize=minsize, maxsize=maxsize, ssl=ssl,
                         parser=parser, loop=loop)
        assert self._address is _NON_DISCOVERED
        self._sentinel = sentinel
        self._service = service
        self._is_master = is_master
        self._discover_timeout = .2

    @property
    def address(self):
        if self._address is _NON_DISCOVERED:
            return
        return self._address

    def get_connection(self, command, args=()):
        if self._address is _NON_DISCOVERED:
            return None, _NON_DISCOVERED
        return super().get_connection(command, args)

    @asyncio.coroutine
    def _create_new_connection(self, address):
        if address is _NON_DISCOVERED:
            # Perform service discovery.
            # Returns Connection or raises error if no service can be found.
            yield from self._do_clear()  # make `clear` blocking

            if self._is_master:
                conn = yield from self._sentinel.discover_master(
                    self._service, timeout=self._sentinel.discover_timeout)
            else:
                conn = yield from self._sentinel.discover_slave(
                    self._service, timeout=self._sentinel.discover_timeout)
            self._address = conn.address
            sentinel_logger.debug("Discoverred new address %r for %s",
                                  conn.address, self._service)
            return conn
        return (yield from super()._create_new_connection(address))

    def _drop_closed(self):
        diff = len(self._pool)
        super()._drop_closed()
        diff -= len(self._pool)
        if diff:
            # closed connections were in pool:
            #   * reset address;
            #   * notify sentinel pool
            sentinel_logger.debug(
                "Dropped %d closed connnection(s); must rediscover", diff)
            self._sentinel._need_rediscover(self._service)

    @asyncio.coroutine
    def acquire(self, command=None, args=()):
        if self._address is _NON_DISCOVERED:
            yield from self.clear()
        return (yield from super().acquire(command, args))

    def release(self, conn):
        was_closed = conn.closed
        super().release(conn)
        # if connection was closed while used and not by release()
        if was_closed:
            sentinel_logger.debug(
                "Released closed connection; must rediscover")
            self._sentinel._need_rediscover(self._service)

    def need_rediscover(self):
        self._address = _NON_DISCOVERED


def make_dict(plain_list):
    it = iter(plain_list)
    return dict(zip(it, it))


class DiscoverError(Exception):
    """Internal errors for masters/slaves discovery."""


class BadState(DiscoverError):
    """Bad master's / slave's state read from sentinel."""


class UnknownService(DiscoverError):
    """Service is not monitored by specific sentinel."""


class RoleMismatch(DiscoverError):
    """Service reported to have other Role."""



from itertools import chain

from aioredis.util import (
    wait_ok,
    wait_convert,
    wait_make_dict,
    _NOTSET,
    PY_35,
    )

if PY_35:
    from aioredis.util import _ScanIterPairs


class HashCommandsMixin:
    """Hash commands mixin.

    For commands details see: http://redis.io/commands#hash
    """

    def hdel(self, key, field, *fields):
        """Delete one or more hash fields."""
        return self.execute(b'HDEL', key, field, *fields)

    def hexists(self, key, field):
        """Determine if hash field exists."""
        fut = self.execute(b'HEXISTS', key, field)
        return wait_convert(fut, bool)

    def hget(self, key, field, *, encoding=_NOTSET):
        """Get the value of a hash field."""
        return self.execute(b'HGET', key, field, encoding=encoding)

    def hgetall(self, key, *, encoding=_NOTSET):
        """Get all the fields and values in a hash."""
        fut = self.execute(b'HGETALL', key, encoding=encoding)
        return wait_make_dict(fut)

    def hincrby(self, key, field, increment=1):
        """Increment the integer value of a hash field by the given number."""
        return self.execute(b'HINCRBY', key, field, increment)

    def hincrbyfloat(self, key, field, increment=1.0):
        """Increment the float value of a hash field by the given number."""
        fut = self.execute(b'HINCRBYFLOAT', key, field, increment)
        return wait_convert(fut, float)

    def hkeys(self, key, *, encoding=_NOTSET):
        """Get all the fields in a hash."""
        return self.execute(b'HKEYS', key, encoding=encoding)

    def hlen(self, key):
        """Get the number of fields in a hash."""
        return self.execute(b'HLEN', key)

    def hmget(self, key, field, *fields, encoding=_NOTSET):
        """Get the values of all the given fields."""
        return self.execute(b'HMGET', key, field, *fields, encoding=encoding)

    def hmset(self, key, field, value, *pairs):
        """Set multiple hash fields to multiple values."""
        if len(pairs) % 2 != 0:
            raise TypeError("length of pairs must be even number")
        return wait_ok(self.execute(b'HMSET', key, field, value, *pairs))

    def hmset_dict(self, key, *args, **kwargs):
        """Set multiple hash fields to multiple values.

        dict can be passed as first positional argument:

        >>> await redis.hmset_dict(
        ...     'key', {'field1': 'value1', 'field2': 'value2'})

        or keyword arguments can be used:

        >>> await redis.hmset_dict(
        ...     'key', field1='value1', field2='value2')

        or dict argument can be mixed with kwargs:

        >>> await redis.hmset_dict(
        ...     'key', {'field1': 'value1'}, field2='value2')

        .. note:: ``dict`` and ``kwargs`` not get mixed into single dictionary,
           if both specified and both have same key(s) -- ``kwargs`` will win:

           >>> await redis.hmset_dict('key', {'foo': 'bar'}, foo='baz')
           >>> await redis.hget('key', 'foo', encoding='utf-8')
           'baz'

        """
        if not args and not kwargs:
            raise TypeError("args or kwargs must be specified")
        pairs = ()
        if len(args) > 1:
            raise TypeError("single positional argument allowed")
        elif len(args) == 1:
            if not isinstance(args[0], dict):
                raise TypeError("args[0] must be dict")
            elif not args[0] and not kwargs:
                raise ValueError("args[0] is empty dict")
            pairs = chain.from_iterable(args[0].items())
        kwargs_pairs = chain.from_iterable(kwargs.items())
        return wait_ok(self.execute(
            b'HMSET', key, *chain(pairs, kwargs_pairs)))

    def hset(self, key, field, value):
        """Set the string value of a hash field."""
        return self.execute(b'HSET', key, field, value)

    def hsetnx(self, key, field, value):
        """Set the value of a hash field, only if the field does not exist."""
        return self.execute(b'HSETNX', key, field, value)

    def hvals(self, key, *, encoding=_NOTSET):
        """Get all the values in a hash."""
        return self.execute(b'HVALS', key, encoding=encoding)

    def hscan(self, key, cursor=0, match=None, count=None):
        """Incrementally iterate hash fields and associated values."""
        args = [key, cursor]
        match is not None and args.extend([b'MATCH', match])
        count is not None and args.extend([b'COUNT', count])
        fut = self.execute(b'HSCAN', *args)
        return wait_convert(fut, lambda obj: (int(obj[0]), obj[1]))

    if PY_35:
        def ihscan(self, key, *, match=None, count=None):
            """Incrementally iterate sorted set items using async for.

            Usage example:

            >>> async for name, val in redis.ihscan(key, match='something*'):
            ...     print('Matched:', name, '->', val)

            """
            return _ScanIterPairs(lambda cur: self.hscan(key, cur,
                                                         match=match,
                                                         count=count))

    def hstrlen(self, key, field):
        """Get the length of the value of a hash field."""
        return self.execute(b'HSTRLEN', key, field)



import pytest


pytestmark = pytest.redis_version(
    2, 8, 9, reason='HyperLogLog works only with redis>=2.8.9')


@pytest.mark.run_loop
def test_pfcount(redis):
    key = 'hll_pfcount'
    other_key = 'some-other-hll'

    # add initial data, cardinality changed so command returns 1
    is_changed = yield from redis.pfadd(key, 'foo', 'bar', 'zap')
    assert is_changed == 1

    # add more data, cardinality not changed so command returns 0
    is_changed = yield from redis.pfadd(key, 'zap', 'zap', 'zap')
    assert is_changed == 0

    # add event more data, cardinality not changed so command returns 0
    is_changed = yield from redis.pfadd(key, 'foo', 'bar')
    assert is_changed == 0

    # check cardinality of one key
    cardinality = yield from redis.pfcount(key)
    assert cardinality == 3

    # create new key (variable) for cardinality estimation
    is_changed = yield from redis.pfadd(other_key, 1, 2, 3)
    assert is_changed == 1

    # check cardinality of multiple keys
    cardinality = yield from redis.pfcount(key, other_key)
    assert cardinality == 6

    with pytest.raises(TypeError):
        yield from redis.pfcount(None)
    with pytest.raises(TypeError):
        yield from redis.pfcount(key, None)
    with pytest.raises(TypeError):
        yield from redis.pfcount(key, key, None)


@pytest.mark.run_loop
def test_pfadd(redis):
    key = 'hll_pfadd'
    values = ['a', 's', 'y', 'n', 'c', 'i', 'o']
    # add initial data, cardinality changed so command returns 1
    is_changed = yield from redis.pfadd(key, *values)
    assert is_changed == 1
    # add event more data, cardinality not changed so command returns 0
    is_changed = yield from redis.pfadd(key, 'i', 'o')
    assert is_changed == 0


@pytest.mark.run_loop
def test_pfadd_wrong_input(redis):
    with pytest.raises(TypeError):
        yield from redis.pfadd(None, 'value')


@pytest.mark.run_loop
def test_pfmerge(redis):
    key = 'hll_asyncio'
    key_other = 'hll_aioredis'

    key_dest = 'hll_aio'

    values = ['a', 's', 'y', 'n', 'c', 'i', 'o']
    values_other = ['a', 'i', 'o', 'r', 'e', 'd', 'i', 's']

    data_set = set(values + values_other)
    cardinality_merged = len(data_set)

    # add initial data, cardinality changed so command returns 1
    yield from redis.pfadd(key, *values)
    yield from redis.pfadd(key_other, *values_other)

    # check cardinality of one key
    cardinality = yield from redis.pfcount(key)
    assert cardinality == len(set(values_other))

    cardinality_other = yield from redis.pfcount(key_other)
    assert cardinality_other == len(set(values_other))

    yield from redis.pfmerge(key_dest, key, key_other)
    cardinality_dest = yield from redis.pfcount(key_dest)
    assert cardinality_dest == cardinality_merged

    with pytest.raises(TypeError):
        yield from redis.pfmerge(None, key)
    with pytest.raises(TypeError):
        yield from redis.pfmerge(key_dest, None)
    with pytest.raises(TypeError):
        yield from redis.pfmerge(key_dest, key, None)


@pytest.mark.run_loop
def test_pfmerge_wrong_input(redis):
    with pytest.raises(TypeError):
        yield from redis.pfmerge(None, 'value')



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def go():
        pool = yield from aioredis.create_pool(
            ('localhost', 6379),
            minsize=5, maxsize=10)
        with (yield from pool) as conn:    # low-level redis connection
            yield from conn.execute('set', 'my-key', 'value')
            val = yield from conn.execute('get', 'my-key')
        print('raw value:', val)
        pool.close()
        yield from pool.wait_closed()    # closing all open connections

    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio
import aioredis

from functools import partial
from timeit import timeit


async def old_pool_or_redis(maxsize=10):
    return await aioredis.create_pool(('localhost', 6379),
                                      maxsize=maxsize)


async def new_redis_pool(maxsize=10):
    return await aioredis.create_redis_pool(('localhost', 6379),
                                            maxsize=maxsize)


async def simple_get_set(pool, val):
    val = 'val:{}'.format(val)
    with await pool as redis:
        assert await redis.set('key', val)
        await redis.get('key', encoding='utf-8')


async def pipeline(pool, val):
    val = 'val:{}'.format(val)
    with await pool as redis:
        f1 = redis.set('key', val)
        f2 = redis.get('key', encoding='utf-8')
        ok, res = await asyncio.gather(f1, f2)


async def transaction(pool, val):
    val = 'val:{}'.format(val)
    with await pool as redis:
        tr = redis.multi_exec()
        tr.set('key', val)
        tr.get('key', encoding='utf-8')
        ok, res = await tr.execute()
        assert ok, ok
        assert res == val


# NOTE: blocking operations (ex: BLPOP) will probably use
#   same connection in new pool so blocking it for long time.
#   This must be fixed.
async def blocking_pop(pool, val):

    async def lpush():
        with await pool as redis:
            await asyncio.sleep(.1)
            await redis.lpush('list-key', 'val')

    async def blpop():
        with await pool as redis:
            res = await redis.blpop('list-key', timeout=1, encoding='utf-8')
            assert res == ['list-key', 'val'], res
    await asyncio.gather(blpop(), lpush())


async def repeat(runnable, n=1):
    done, pending = await asyncio.wait([asyncio.ensure_future(runnable(i))
                                        for i in range(n)])
    assert not pending
    done = [f.exception() for f in done]
    a = b = 0
    for rv in done:
        if rv is None:
            a += 1
        else:
            b += 1
    print("success/failure: {}/{}".format(a, b))


if __name__ == '__main__':
    run = asyncio.get_event_loop().run_until_complete

    if hasattr(aioredis, 'create_redis_pool'):
        print("Testing new Redis pool")
        pool = run(new_redis_pool(maxsize=5))
    else:
        print("Testing old Pool or Redises")
        pool = run(old_pool_or_redis(maxsize=5))

    async def cleanup():
        with await pool as client:
            await client.flushdb()
    run(cleanup())
    try:
        print("running simple get/set test…", end=' ')
        t = timeit(lambda: run(repeat(partial(simple_get_set, pool), n=100)),
                   number=1)
        print("took:", t)
        print("running pipeline test…", end=' ')
        t = timeit(lambda: run(repeat(partial(pipeline, pool), n=100)),
                   number=1)
        print("took:", t)
        print("running multi-exec test…", end=' ')
        t = timeit(lambda: run(repeat(partial(transaction, pool), n=100)),
                   number=1)
        print("took:", t)
        print("running blpop test…", end=' ')
        t = timeit(lambda: run(repeat(partial(blocking_pop, pool), n=100)),
                   number=1)
        print("took:", t)
    finally:
        pool.close()
        run(pool.wait_closed())



import asyncio
import aioredis


def main():
    loop = asyncio.get_event_loop()

    @asyncio.coroutine
    def reader(ch):
        while (yield from ch.wait_message()):
            msg = yield from ch.get_json()
            print("Got Message:", msg)

    @asyncio.coroutine
    def go():
        pub = yield from aioredis.create_redis(
            ('localhost', 6379))
        sub = yield from aioredis.create_redis(
            ('localhost', 6379))
        res = yield from sub.subscribe('chan:1')
        ch1 = res[0]

        tsk = asyncio.async(reader(ch1))

        res = yield from pub.publish_json('chan:1', ["Hello", "world"])
        assert res == 1

        yield from sub.unsubscribe('chan:1')
        yield from tsk
        sub.close()
        pub.close()

    loop.run_until_complete(go())


if __name__ == '__main__':
    main()



import asyncio
import aioredis


@asyncio.coroutine
def pubsub():
    sub = yield from aioredis.create_redis(
         ('localhost', 6379))

    ch1, ch2 = yield from sub.subscribe('channel:1', 'channel:2')
    assert isinstance(ch1, aioredis.Channel)
    assert isinstance(ch2, aioredis.Channel)

    @asyncio.coroutine
    def async_reader(channel):
        while (yield from channel.wait_message()):
            msg = yield from channel.get(encoding='utf-8')
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

    tsk1 = asyncio.async(async_reader(ch1))

    # Or alternatively:

    @asyncio.coroutine
    def async_reader2(channel):
        while True:
            msg = yield from channel.get(encoding='utf-8')
            if msg is None:
                break
            # ... process message ...
            print("message in {}: {}".format(channel.name, msg))

    tsk2 = asyncio.async(async_reader2(ch2))

    # Publish messages and terminate
    pub = yield from aioredis.create_redis(
        ('localhost', 6379))
    while True:
        channels = yield from pub.pubsub_channels('channel:*')
        if len(channels) == 2:
            break

    for msg in ("Hello", ",", "world!"):
        for ch in ('channel:1', 'channel:2'):
            yield from pub.publish(ch, msg)
    pub.close()
    sub.close()
    yield from asyncio.sleep(0)
    yield from pub.wait_closed()
    yield from sub.wait_closed()
    yield from asyncio.gather(tsk1, tsk2)


if __name__ == '__main__':
    import os
    if 'redis_version:2.6' not in os.environ.get('REDIS_VERSION', ''):
        loop = asyncio.get_event_loop()
        loop.run_until_complete(pubsub())



import re
import os.path
import sys
import platform
from setuptools import setup, find_packages


install_requires = ['async-timeout']
if platform.python_implementation() == 'CPython':
    install_requires.append('hiredis')

PY_VER = sys.version_info

if PY_VER >= (3, 4):
    pass
elif PY_VER >= (3, 3):
    install_requires.append('asyncio')
else:
    raise RuntimeError("aioredis doesn't support Python version prior 3.3")


def read(*parts):
    with open(os.path.join(*parts), 'rt') as f:
        return f.read().strip()


def read_version():
    regexp = re.compile(r"^__version__\W*=\W*'([\d.abrc]+)'")
    init_py = os.path.join(os.path.dirname(__file__),
                           'aioredis', '__init__.py')
    with open(init_py) as f:
        for line in f:
            match = regexp.match(line)
            if match is not None:
                return match.group(1)
        else:
            raise RuntimeError('Cannot find version in aioredis/__init__.py')


classifiers = [
    'License :: OSI Approved :: MIT License',
    'Development Status :: 4 - Beta',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Operating System :: POSIX',
    'Environment :: Web Environment',
    'Intended Audience :: Developers',
    'Topic :: Software Development',
    'Topic :: Software Development :: Libraries',
    'Framework :: AsyncIO',
]

setup(name='aioredis',
      version=read_version(),
      description=("asyncio (PEP 3156) Redis support"),
      long_description="\n\n".join((read('README.rst'), read('CHANGES.txt'))),
      classifiers=classifiers,
      platforms=["POSIX"],
      author="Alexey Popravka",
      author_email="alexey.popravka@horsedevel.com",
      url="https://github.com/aio-libs/aioredis",
      license="MIT",
      packages=find_packages(exclude=["tests"]),
      install_requires=install_requires,
      include_package_data=True,
      )



from .commands import RedisSentinel, create_sentinel
from .pool import SentinelPool, create_sentinel_pool

__all__ = [
    "create_sentinel",
    "create_sentinel_pool",
    "RedisSentinel",
    "SentinelPool",
]



import asyncio
import pytest

from aioredis.pubsub import Receiver


@pytest.mark.run_loop
async def test_pubsub_receiver_iter(create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)
    pub = await create_redis(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)

    async def coro(mpsc):
        lst = []
        async for msg in mpsc.iter():
            lst.append(msg)
        return lst

    tsk = asyncio.ensure_future(coro(mpsc), loop=loop)
    snd1, = await sub.subscribe(mpsc.channel('chan:1'))
    snd2, = await sub.subscribe(mpsc.channel('chan:2'))
    snd3, = await sub.psubscribe(mpsc.pattern('chan:*'))

    subscribers = await pub.publish_json('chan:1', {'Hello': 'World'})
    assert subscribers > 1
    subscribers = await pub.publish_json('chan:2', ['message'])
    assert subscribers > 1
    loop.call_later(0, mpsc.stop)
    # await asyncio.sleep(0, loop=loop)
    assert await tsk == [
        (snd1, b'{"Hello": "World"}'),
        (snd3, (b'chan:1', b'{"Hello": "World"}')),
        (snd2, b'["message"]'),
        (snd3, (b'chan:2', b'["message"]')),
        ]
    assert not mpsc.is_active


@pytest.mark.run_loop(timeout=5)
async def test_pubsub_receiver_call_stop_with_empty_queue(
        create_redis, server, loop):
    sub = await create_redis(server.tcp_address, loop=loop)

    mpsc = Receiver(loop=loop)

    # FIXME: currently at least one subscriber is needed
    snd1, = await sub.subscribe(mpsc.channel('chan:1'))

    now = loop.time()
    loop.call_later(.5, mpsc.stop)
    async for i in mpsc.iter():  # noqa (flake8 bug with async for)
        assert False, "StopAsyncIteration not raised"
    dt = loop.time() - now
    assert dt <= 1.5
    assert not mpsc.is_active



import pytest

from aioredis.util import encode_command


def test_encode_bytes():
    res = encode_command(b'Hello')
    assert res == b'*1\r\n$5\r\nHello\r\n'

    res = encode_command(b'Hello', b'World')
    assert res == b'*2\r\n$5\r\nHello\r\n$5\r\nWorld\r\n'

    res = encode_command(b'\0')
    assert res == b'*1\r\n$1\r\n\0\r\n'

    res = encode_command(bytearray(b'Hello\r\n'))
    assert res == b'*1\r\n$7\r\nHello\r\n\r\n'


def test_encode_str():
    res = encode_command('Hello')
    assert res == b'*1\r\n$5\r\nHello\r\n'

    res = encode_command('Hello', 'world')
    assert res == b'*2\r\n$5\r\nHello\r\n$5\r\nworld\r\n'


def test_encode_int():
    res = encode_command(1)
    assert res == b'*1\r\n$1\r\n1\r\n'

    res = encode_command(-1)
    assert res == b'*1\r\n$2\r\n-1\r\n'


def test_encode_float():
    res = encode_command(1.0)
    assert res == b'*1\r\n$3\r\n1.0\r\n'

    res = encode_command(-1.0)
    assert res == b'*1\r\n$4\r\n-1.0\r\n'


def test_encode_empty():
    res = encode_command()
    assert res == b'*0\r\n'


def test_encode_errors():
    with pytest.raises(TypeError):
        encode_command(dict())
    with pytest.raises(TypeError):
        encode_command(list())
    with pytest.raises(TypeError):
        encode_command(None)



import pytest


@pytest.mark.run_loop
async def test_await(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    with await pool as conn:
        msg = await conn.execute('echo', 'hello')
        assert msg == b'hello'


@pytest.mark.run_loop
async def test_async_with(create_pool, server, loop):
    pool = await create_pool(
        server.tcp_address,
        minsize=10, loop=loop)

    async with pool.get() as conn:
        msg = await conn.execute('echo', 'hello')
        assert msg == b'hello'



import pytest
import asyncio
import sys
from unittest import mock

from unittest.mock import patch

from aioredis.util import async_task
from aioredis import (
    ConnectionClosedError,
    ProtocolError,
    RedisConnection,
    RedisError,
    ReplyError,
    Channel,
    )


@pytest.mark.run_loop
def test_connect_tcp(request, create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)
    assert conn.db == 0
    assert isinstance(conn.address, tuple)
    assert conn.address[0] in ('127.0.0.1', '::1')
    assert conn.address[1] == server.tcp_address.port
    assert str(conn) == '<RedisConnection [db:0]>'

    conn = yield from create_connection(
        ['localhost', server.tcp_address.port], loop=loop)
    assert conn.db == 0
    assert isinstance(conn.address, tuple)
    assert conn.address[0] in ('127.0.0.1', '::1')
    assert conn.address[1] == server.tcp_address.port
    assert str(conn) == '<RedisConnection [db:0]>'


@pytest.mark.run_loop
def test_connect_inject_connection_cls(
        request,
        create_connection,
        loop,
        server):

    class MyConnection(RedisConnection):
        pass

    conn = yield from create_connection(
        server.tcp_address, loop=loop, connection_cls=MyConnection)

    assert isinstance(conn, MyConnection)


@pytest.mark.run_loop
def test_connect_inject_connection_cls_invalid(
        request,
        create_connection,
        loop,
        server):

    with pytest.raises(AssertionError):
        yield from create_connection(
            server.tcp_address, loop=loop, connection_cls=type)


@pytest.mark.run_loop
def test_connect_tcp_timeout(request, create_connection, loop, server):
    with patch('aioredis.connection.asyncio.open_connection') as\
            open_conn_mock:
        open_conn_mock.side_effect = lambda *a, **kw: asyncio.sleep(0.2,
                                                                    loop=loop)
        with pytest.raises(asyncio.TimeoutError):
            yield from create_connection(
                server.tcp_address, loop=loop, timeout=0.1)


@pytest.mark.run_loop
def test_connect_tcp_invalid_timeout(request, create_connection, loop, server):
    with pytest.raises(ValueError):
        yield from create_connection(
            server.tcp_address, loop=loop, timeout=0)


@pytest.mark.run_loop
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="No unixsocket on Windows")
def test_connect_unixsocket(create_connection, loop, server):
    conn = yield from create_connection(
        server.unixsocket, db=0, loop=loop)
    assert conn.db == 0
    assert conn.address == server.unixsocket
    assert str(conn) == '<RedisConnection [db:0]>'


@pytest.mark.run_loop
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="No unixsocket on Windows")
def test_connect_unixsocket_timeout(create_connection, loop, server):
    with patch('aioredis.connection.asyncio.open_unix_connection') as\
            open_conn_mock:
        open_conn_mock.side_effect = lambda *a, **kw: asyncio.sleep(0.2,
                                                                    loop=loop)
        with pytest.raises(asyncio.TimeoutError):
            yield from create_connection(
                server.unixsocket, db=0, loop=loop, timeout=0.1)


def test_global_loop(create_connection, loop, server):
    asyncio.set_event_loop(loop)

    conn = loop.run_until_complete(create_connection(
        server.tcp_address, db=0))
    assert conn.db == 0
    assert conn._loop is loop


@pytest.mark.run_loop
def test_select_db(create_connection, loop, server):
    address = server.tcp_address
    conn = yield from create_connection(address, loop=loop)
    assert conn.db == 0

    with pytest.raises(ValueError):
        yield from create_connection(address, db=-1, loop=loop)
    with pytest.raises(TypeError):
        yield from create_connection(address, db=1.0, loop=loop)
    with pytest.raises(TypeError):
        yield from create_connection(
            address, db='bad value', loop=loop)
    with pytest.raises(TypeError):
        conn = yield from create_connection(
            address, db=None, loop=loop)
        yield from conn.select(None)
    with pytest.raises(ReplyError):
        yield from create_connection(
            address, db=100000, loop=loop)

    yield from conn.select(1)
    assert conn.db == 1
    yield from conn.select(2)
    assert conn.db == 2
    yield from conn.execute('select', 0)
    assert conn.db == 0
    yield from conn.execute(b'select', 1)
    assert conn.db == 1


@pytest.mark.run_loop
def test_protocol_error(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)

    reader = conn._reader

    with pytest.raises(ProtocolError):
        reader.feed_data(b'not good redis protocol response')
        yield from conn.select(1)

    assert len(conn._waiters) == 0


def test_close_connection__tcp(create_connection, loop, server):
    conn = loop.run_until_complete(create_connection(
        server.tcp_address, loop=loop))
    conn.close()
    with pytest.raises(ConnectionClosedError):
        loop.run_until_complete(conn.select(1))

    conn = loop.run_until_complete(create_connection(
        server.tcp_address, loop=loop))
    conn.close()
    fut = None
    with pytest.raises(ConnectionClosedError):
        fut = conn.select(1)
    assert fut is None

    conn = loop.run_until_complete(create_connection(
        server.tcp_address, loop=loop))
    conn.close()
    with pytest.raises(ConnectionClosedError):
        conn.execute_pubsub('subscribe', 'channel:1')


@pytest.mark.run_loop
@pytest.mark.skipif(sys.platform == 'win32',
                    reason="No unixsocket on Windows")
def test_close_connection__socket(create_connection, loop, server):
    conn = yield from create_connection(
        server.unixsocket, loop=loop)
    conn.close()
    with pytest.raises(ConnectionClosedError):
        yield from conn.select(1)

    conn = yield from create_connection(
        server.unixsocket, loop=loop)
    conn.close()
    with pytest.raises(ConnectionClosedError):
        yield from conn.execute_pubsub('subscribe', 'channel:1')


@pytest.mark.run_loop
def test_closed_connection_with_none_reader(create_connection, loop, server):
    address = server.tcp_address
    conn = yield from create_connection(address, loop=loop)
    stored_reader = conn._reader
    conn._reader = None
    with pytest.raises(ConnectionClosedError):
        yield from conn.execute('blpop', 'test', 0)
    conn._reader = stored_reader
    conn.close()

    conn = yield from create_connection(address, loop=loop)
    stored_reader = conn._reader
    conn._reader = None
    with pytest.raises(ConnectionClosedError):
        yield from conn.execute_pubsub('subscribe', 'channel:1')
    conn._reader = stored_reader
    conn.close()


@pytest.mark.run_loop
def test_wait_closed(create_connection, loop, server):
    address = server.tcp_address
    conn = yield from create_connection(address, loop=loop)
    reader_task = conn._reader_task
    conn.close()
    assert not reader_task.done()
    yield from conn.wait_closed()
    assert reader_task.done()


@pytest.mark.run_loop
def test_cancel_wait_closed(create_connection, loop, server):
    # Regression test: Don't throw error if wait_closed() is cancelled.
    address = server.tcp_address
    conn = yield from create_connection(address, loop=loop)
    reader_task = conn._reader_task
    conn.close()
    task = async_task(conn.wait_closed(), loop=loop)

    # Make sure the task is cancelled
    # after it has been started by the loop.
    loop.call_soon(task.cancel)

    yield from conn.wait_closed()
    assert reader_task.done()


@pytest.mark.run_loop
def test_auth(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)

    res = yield from conn.execute('CONFIG', 'SET', 'requirepass', 'pass')
    assert res == b'OK'

    conn2 = yield from create_connection(
        server.tcp_address, loop=loop)

    with pytest.raises(ReplyError):
        yield from conn2.select(1)

    res = yield from conn2.auth('pass')
    assert res is True
    res = yield from conn2.select(1)
    assert res is True

    conn3 = yield from create_connection(
        server.tcp_address, password='pass', loop=loop)

    res = yield from conn3.select(1)
    assert res is True

    res = yield from conn2.execute('CONFIG', 'SET', 'requirepass', '')
    assert res == b'OK'


@pytest.mark.run_loop
def test_decoding(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, encoding='utf-8', loop=loop)
    assert conn.encoding == 'utf-8'
    res = yield from conn.execute('set', '{prefix}:key1', 'value')
    assert res == 'OK'
    res = yield from conn.execute('get', '{prefix}:key1')
    assert res == 'value'

    res = yield from conn.execute('set', '{prefix}:key1', b'bin-value')
    assert res == 'OK'
    res = yield from conn.execute('get', '{prefix}:key1')
    assert res == 'bin-value'

    res = yield from conn.execute('get', '{prefix}:key1', encoding='ascii')
    assert res == 'bin-value'
    res = yield from conn.execute('get', '{prefix}:key1', encoding=None)
    assert res == b'bin-value'

    with pytest.raises(UnicodeDecodeError):
        yield from conn.execute('set', '{prefix}:key1', 'значение')
        yield from conn.execute('get', '{prefix}:key1', encoding='ascii')

    conn2 = yield from create_connection(
        server.tcp_address, loop=loop)
    res = yield from conn2.execute('get', '{prefix}:key1', encoding='utf-8')
    assert res == 'значение'


@pytest.mark.run_loop
def test_execute_exceptions(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)
    with pytest.raises(TypeError):
        yield from conn.execute(None)
    with pytest.raises(TypeError):
        yield from conn.execute("ECHO", None)
    with pytest.raises(TypeError):
        yield from conn.execute("GET", ('a', 'b'))
    assert len(conn._waiters) == 0


@pytest.mark.run_loop
def test_subscribe_unsubscribe(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)

    assert conn.in_pubsub == 0

    res = yield from conn.execute('subscribe', 'chan:1')
    assert res == [[b'subscribe', b'chan:1', 1]]

    assert conn.in_pubsub == 1

    res = yield from conn.execute('unsubscribe', 'chan:1')
    assert res == [[b'unsubscribe', b'chan:1', 0]]
    assert conn.in_pubsub == 0

    res = yield from conn.execute('subscribe', 'chan:1', 'chan:2')
    assert res == [[b'subscribe', b'chan:1', 1],
                   [b'subscribe', b'chan:2', 2],
                   ]
    assert conn.in_pubsub == 2

    res = yield from conn.execute('unsubscribe', 'non-existent')
    assert res == [[b'unsubscribe', b'non-existent', 2]]
    assert conn.in_pubsub == 2

    res = yield from conn.execute('unsubscribe', 'chan:1')
    assert res == [[b'unsubscribe', b'chan:1', 1]]
    assert conn.in_pubsub == 1


@pytest.mark.run_loop
def test_psubscribe_punsubscribe(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)
    res = yield from conn.execute('psubscribe', 'chan:*')
    assert res == [[b'psubscribe', b'chan:*', 1]]
    assert conn.in_pubsub == 1


@pytest.mark.run_loop
def test_bad_command_in_pubsub(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop)

    res = yield from conn.execute('subscribe', 'chan:1')
    assert res == [[b'subscribe', b'chan:1', 1]]

    msg = "Connection in SUBSCRIBE mode"
    with pytest.raises(RedisError, match=msg):
        yield from conn.execute('select', 1)
    with pytest.raises(RedisError, match=msg):
        conn.execute('get')


@pytest.mark.run_loop
def test_pubsub_messages(create_connection, loop, server):
    sub = yield from create_connection(
        server.tcp_address, loop=loop)
    pub = yield from create_connection(
        server.tcp_address, loop=loop)
    res = yield from sub.execute('subscribe', 'chan:1')
    assert res == [[b'subscribe', b'chan:1', 1]]

    assert b'chan:1' in sub.pubsub_channels
    chan = sub.pubsub_channels[b'chan:1']
    assert str(chan) == "<Channel name:b'chan:1', is_pattern:False, qsize:0>"
    assert chan.name == b'chan:1'
    assert chan.is_active is True

    res = yield from pub.execute('publish', 'chan:1', 'Hello!')
    assert res == 1
    msg = yield from chan.get()
    assert msg == b'Hello!'

    res = yield from sub.execute('psubscribe', 'chan:*')
    assert res == [[b'psubscribe', b'chan:*', 2]]
    assert b'chan:*' in sub.pubsub_patterns
    chan2 = sub.pubsub_patterns[b'chan:*']
    assert chan2.name == b'chan:*'
    assert chan2.is_active is True

    res = yield from pub.execute('publish', 'chan:1', 'Hello!')
    assert res == 2

    msg = yield from chan.get()
    assert msg == b'Hello!'
    dest_chan, msg = yield from chan2.get()
    assert dest_chan == b'chan:1'
    assert msg == b'Hello!'


@pytest.mark.run_loop
def test_multiple_subscribe_unsubscribe(create_connection, loop, server):
    sub = yield from create_connection(server.tcp_address, loop=loop)

    res = yield from sub.execute_pubsub('subscribe', 'chan:1')
    ch = sub.pubsub_channels['chan:1']
    assert res == [[b'subscribe', b'chan:1', 1]]
    res = yield from sub.execute_pubsub('subscribe', b'chan:1')
    assert res == [[b'subscribe', b'chan:1', 1]]
    assert ch is sub.pubsub_channels['chan:1']
    res = yield from sub.execute_pubsub('subscribe', ch)
    assert res == [[b'subscribe', b'chan:1', 1]]
    assert ch is sub.pubsub_channels['chan:1']

    res = yield from sub.execute_pubsub('unsubscribe', 'chan:1')
    assert res == [[b'unsubscribe', b'chan:1', 0]]
    res = yield from sub.execute_pubsub('unsubscribe', 'chan:1')
    assert res == [[b'unsubscribe', b'chan:1', 0]]

    res = yield from sub.execute_pubsub('psubscribe', 'chan:*')
    assert res == [[b'psubscribe', b'chan:*', 1]]
    res = yield from sub.execute_pubsub('psubscribe', 'chan:*')
    assert res == [[b'psubscribe', b'chan:*', 1]]

    res = yield from sub.execute_pubsub('punsubscribe', 'chan:*')
    assert res == [[b'punsubscribe', b'chan:*', 0]]
    res = yield from sub.execute_pubsub('punsubscribe', 'chan:*')
    assert res == [[b'punsubscribe', b'chan:*', 0]]


@pytest.mark.run_loop
def test_execute_pubsub_errors(create_connection, loop, server):
    sub = yield from create_connection(
        server.tcp_address, loop=loop)

    with pytest.raises(TypeError):
        sub.execute_pubsub('subscribe', "chan:1", None)
    with pytest.raises(TypeError):
        sub.execute_pubsub('subscribe')
    with pytest.raises(ValueError):
        sub.execute_pubsub(
            'subscribe',
            Channel('chan:1', is_pattern=True, loop=loop))
    with pytest.raises(ValueError):
        sub.execute_pubsub(
            'unsubscribe',
            Channel('chan:1', is_pattern=True, loop=loop))
    with pytest.raises(ValueError):
        sub.execute_pubsub(
            'psubscribe',
            Channel('chan:1', is_pattern=False, loop=loop))
    with pytest.raises(ValueError):
        sub.execute_pubsub(
            'punsubscribe',
            Channel('chan:1', is_pattern=False, loop=loop))


@pytest.mark.run_loop
def test_multi_exec(create_connection, loop, server):
    conn = yield from create_connection(server.tcp_address, loop=loop)

    ok = yield from conn.execute('set', 'foo', 'bar')
    assert ok == b'OK'

    ok = yield from conn.execute("MULTI")
    assert ok == b'OK'
    queued = yield from conn.execute('getset', 'foo', 'baz')
    assert queued == b'QUEUED'
    res = yield from conn.execute("EXEC")
    assert res == [b'bar']

    ok = yield from conn.execute("MULTI")
    assert ok == b'OK'
    queued = yield from conn.execute('getset', 'foo', 'baz')
    assert queued == b'QUEUED'
    res = yield from conn.execute("DISCARD")
    assert res == b'OK'


@pytest.mark.run_loop
def test_multi_exec__enc(create_connection, loop, server):
    conn = yield from create_connection(
        server.tcp_address, loop=loop, encoding='utf-8')

    ok = yield from conn.execute('set', 'foo', 'bar')
    assert ok == 'OK'

    ok = yield from conn.execute("MULTI")
    assert ok == 'OK'
    queued = yield from conn.execute('getset', 'foo', 'baz')
    assert queued == 'QUEUED'
    res = yield from conn.execute("EXEC")
    assert res == ['bar']

    ok = yield from conn.execute("MULTI")
    assert ok == 'OK'
    queued = yield from conn.execute('getset', 'foo', 'baz')
    assert queued == 'QUEUED'
    res = yield from conn.execute("DISCARD")
    assert res == 'OK'


@pytest.mark.run_loop
def test_connection_parser_argument(create_connection, server, loop):
    klass = mock.MagicMock()
    klass.return_value = reader = mock.Mock()
    conn = yield from create_connection(server.tcp_address,
                                        parser=klass, loop=loop)

    def ret_once(*args, _called=[False]):
        if not _called[0]:
            _called[0] = True
            return b'xxx'
        return False

    reader.gets.side_effect = ret_once
    assert b'xxx' == (yield from conn.execute('ping'))

    assert klass.mock_calls == [
        mock.call(protocolError=ProtocolError, replyError=ReplyError),
        mock.call().feed(b'+PONG\r\n'),
        mock.call().gets(),
        mock.call().gets(),
    ]



"""The module provides connection and connections pool interfaces.

These are intended to be used for implementing custom connection managers.
"""
import abc
import asyncio
try:
    from abc import ABC
except ImportError:
    class ABC(metaclass=abc.ABCMeta):
        pass


__all__ = [
    'AbcConnection',
    'AbcPool',
    'AbcChannel',
]


class AbcConnection(ABC):
    """Abstract connection interface."""

    @abc.abstractmethod
    def execute(self, command, *args, **kwargs):
        """Execute redis command."""

    @abc.abstractmethod
    def execute_pubsub(self, command, *args, **kwargs):
        """Execute Redis (p)subscribe/(p)unsubscribe commands."""

    @abc.abstractmethod
    def close(self):
        """Perform connection(s) close and resources cleanup."""

    @asyncio.coroutine
    @abc.abstractmethod
    def wait_closed(self):
        """
        Coroutine waiting until all resources are closed/released/cleaned up.
        """

    @property
    @abc.abstractmethod
    def closed(self):
        """Flag indicating if connection is closing or already closed."""

    @property
    @abc.abstractmethod
    def db(self):
        """Current selected DB index."""

    @property
    @abc.abstractmethod
    def encoding(self):
        """Current set connection codec."""

    @property
    @abc.abstractmethod
    def in_pubsub(self):
        """Returns number of subscribed channels.

        Can be tested as bool indicating Pub/Sub mode state.
        """

    @property
    @abc.abstractmethod
    def pubsub_channels(self):
        """Read-only channels dict."""

    @property
    @abc.abstractmethod
    def pubsub_patterns(self):
        """Read-only patterns dict."""

    @property
    @abc.abstractmethod
    def address(self):
        """Connection address."""


class AbcPool(AbcConnection):
    """Abstract connections pool interface.

    Inherited from AbcConnection so both have common interface
    for executing Redis commands.
    """

    @abc.abstractmethod
    def get_connection(self):   # TODO: arguments
        """
        Gets free connection from pool in a sync way.

        If no connection available — returns None.
        """

    @asyncio.coroutine
    @abc.abstractmethod
    def acquire(self):  # TODO: arguments
        """Acquires connection from pool."""

    @abc.abstractmethod
    def release(self, conn):  # TODO: arguments
        """Releases connection to pool.

        :param AbcConnection conn: Owned connection to be released.
        """

    @property
    @abc.abstractmethod
    def address(self):
        """Connection address or None."""


class AbcChannel(ABC):
    """Abstract Pub/Sub Channel interface."""

    @property
    @abc.abstractmethod
    def name(self):
        """Encoded channel name or pattern."""

    @property
    @abc.abstractmethod
    def is_pattern(self):
        """Boolean flag indicating if channel is pattern channel."""

    @property
    @abc.abstractmethod
    def is_active(self):
        """Flag indicating that channel has unreceived messages
        and not marked as closed."""

    @asyncio.coroutine
    @abc.abstractmethod
    def get(self):
        """Wait and return new message.

        Will raise ``ChannelClosedError`` if channel is not active.
        """

    # wait_message is not required; details of implementation
    # @abc.abstractmethod
    # def wait_message(self):
    #     pass

    @abc.abstractmethod
    def put_nowait(self, data):
        """Send data to channel.

        Called by RedisConnection when new message received.
        For pattern subscriptions data will be a tuple of
        channel name and message itself.
        """

    @abc.abstractmethod
    def close(self):
        """Marks Channel as closed, no more messages will be sent to it.

        Called by RedisConnection when channel is unsubscribed
        or connection is closed.
        """



import asyncio
import aioredis


async def main():
    redis = await aioredis.create_redis(
        ('localhost', 6379))

    async def transaction():
        tr = redis.multi_exec()
        future1 = tr.set('foo', '123')
        future2 = tr.set('bar', '321')
        result = await tr.execute()
        assert result == await asyncio.gather(future1, future2)
        return result

    await transaction()
    redis.close()
    await redis.wait_closed()


if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main())



import pytest
import asyncio

from aioredis import ConnectionClosedError, ReplyError
from aioredis.pool import ConnectionsPool
from aioredis import Redis


@pytest.mark.run_loop
def test_repr(create_redis, loop, server):
    redis = yield from create_redis(
        server.tcp_address, db=1, loop=loop)
    assert repr(redis) in {
        '<Redis <RedisConnection [db:1]>>',
        '<Redis <ConnectionsPool [db:1, size:[1:10], free:1]>>',
        }

    redis = yield from create_redis(
        server.tcp_address, db=0, loop=loop)
    assert repr(redis) in {
        '<Redis <RedisConnection [db:0]>>',
        '<Redis <ConnectionsPool [db:0, size:[1:10], free:1]>>',
        }


@pytest.mark.run_loop
def test_auth(redis):
    expected_message = "ERR Client sent AUTH, but no password is set"
    with pytest.raises(ReplyError, match=expected_message):
        yield from redis.auth('')


@pytest.mark.run_loop
def test_echo(redis):
    resp = yield from redis.echo('ECHO')
    assert resp == b'ECHO'

    with pytest.raises(TypeError):
        yield from redis.echo(None)


@pytest.mark.run_loop
def test_ping(redis):
    resp = yield from redis.ping()
    assert resp == b'PONG'


@pytest.mark.run_loop
def test_quit(redis, loop):
    try:
        assert b'OK' == (yield from redis.quit())
    except asyncio.CancelledError:
        pass

    if not isinstance(redis.connection, ConnectionsPool):
        expected = (asyncio.CancelledError, ConnectionClosedError)
        # reader task may not yet been cancelled and _do_close not called
        #   so the CancelledError may be raised here
        with pytest.raises(expected):
            yield from redis.ping()

        # wait one loop iteration until it get surely closed
        yield from asyncio.sleep(0, loop=loop)
        assert redis.connection.closed

        with pytest.raises(ConnectionClosedError):
            yield from redis.ping()


@pytest.mark.run_loop
def test_select(redis):
    assert redis.db == 0

    resp = yield from redis.select(1)
    assert resp is True
    assert redis.db == 1
    assert redis.connection.db == 1


@pytest.mark.run_loop
def test_encoding(create_redis, loop, server):
    redis = yield from create_redis(
        server.tcp_address,
        db=1, encoding='utf-8',
        loop=loop)
    assert redis.encoding == 'utf-8'


@pytest.mark.run_loop
def test_yield_from_backwards_compatability(create_redis, server, loop):
    redis = yield from create_redis(server.tcp_address, loop=loop)

    assert isinstance(redis, Redis)
    with pytest.warns(UserWarning):
        with (yield from redis) as client:
            assert isinstance(client, Redis)
            assert client is redis
            assert (yield from client.ping())



import pytest


@pytest.redis_version(2, 8, 0, reason='SCAN is available since redis>=2.8.0')
@pytest.mark.run_loop
async def test_iscan(redis):
    full = set()
    foo = set()
    bar = set()
    for i in range(1, 11):
        is_bar = i % 3
        foo_or_bar = 'bar' if is_bar else 'foo'
        key = 'key:scan:{}:{}'.format(foo_or_bar, i).encode('utf-8')
        full.add(key)
        if is_bar:
            bar.add(key)
        else:
            foo.add(key)
        assert await redis.set(key, i) is True

    async def coro(cmd):
        lst = []
        async for i in cmd:
            lst.append(i)
        return lst

    ret = await coro(redis.iscan())
    assert len(ret) >= 10

    ret = await coro(redis.iscan(match='key:scan:*'))
    assert 10 == len(ret)
    assert set(ret) == full

    ret = await coro(redis.iscan(match='key:scan:foo*'))
    assert set(ret) == foo

    ret = await coro(redis.iscan(match='key:scan:bar*'))
    assert set(ret) == bar

    # SCAN family functions do not guarantee that the number of
    # elements returned per call are in a given range. So here
    # just dummy test, that *count* argument does not break something

    ret = await coro(redis.iscan(match='key:scan:*', count=2))
    assert 10 == len(ret)
    assert set(ret) == full
